#pragma once

#include <cmath>

#include <amdis/LinearAlgebra.hpp>
#include <amdis/OperatorList.hpp>
#include <amdis/common/FlatMatrix.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/linearalgebra/SymmetryStructure.hpp>
#include <amdis/typetree/TreePath.hpp>

namespace AMDiS
{
  /**
   * Basis implementation of DOFMatrix, i.e. a sparse matrix storing all the
   * assembled Operators indexed with DOF indices. The matrix data is associated
   * to a row and column global basis.
   *
   * \tparam RB  Basis of the matrix rows
   * \tparam CB  Basis of matrix columns
   * \tparam T   Coefficient type to store in the matrix
   **/
  template <class RB, class CB, class T = double>
  class BiLinearForm
      : public MatrixBase<RB,CB,MatrixBackend<BackendTraits<RB,T>>>
  {
    using Self = BiLinearForm;
    using Traits = BackendTraits<RB,T>;
    using Backend = MatrixBackend<Traits>;
    using Super = MatrixBase<RB,CB,Backend>;

  public:
    /// The type of the finite element space / basis of the row
    using RowBasis = RB;
    using RowLocalView = typename RowBasis::LocalView;

    /// The type of the finite element space / basis of the column
    using ColBasis = CB;
    using ColLocalView = typename ColBasis::LocalView;

    /// The type of the elements of the DOFVector
    using CoefficientType = typename Traits::CoefficientType;

    /// The type of the matrix filled on an element with local contributions
    using ElementMatrix = FlatMatrix<CoefficientType>;

  public:
    /// Constructor. Wraps the reference into a non-destroying shared_ptr or moves the basis into
    /// a new shared_ptr.
    template <class RB_, class CB_>
    BiLinearForm(RB_&& rowBasis, CB_&& colBasis)
      : Super(FWD(rowBasis), FWD(colBasis))
    {
      operators_.init(*Super::rowBasis(), *Super::colBasis());
    }

    /// Prepare the matrix for insertion. Clears all the entries.
    void init(SymmetryStructure symmetry = SymmetryStructure::unknown)
    {
      Super::init(symmetry);

      auto const rowSize = this->rowBasis()->localView().maxSize();
      auto const colSize = this->colBasis()->localView().maxSize();
      elementMatrix_.resize(rowSize, colSize);
    }

    /// \brief Associate a local operator with this BiLinearForm
    /**
     * Stores an operator in a list that gets assembled during a call to \ref assemble().
     * The operator may be assigned to a specific context, i.e. either an element
     * operator, an intersection operator, or a boundary operator.
     * The \p row and \p col tree paths specify the sub-basis for test and trial
     * functions the operator is applied to.
     *
     * \tparam ContextTag  One of \ref tag::element_operator, \ref tag::intersection_operator
     *                     or \ref tag::boundary_operator indicating where to assemble
     *                     this operator.
     * \tparam Expr        An pre-operator that can be bound to a gridView, or a valid
     *                     GridOperator.
     * \tparam row         A tree-path for the RowBasis
     * \tparam col         A tree-path for the ColBasis
     *
     * [[expects: row is valid tree-path in RowBasis]]
     * [[expects: col is valid tree-path in ColBasis]]
     * @{
     **/
    template <class ContextTag, class Expr, class Row, class Col>
    void addOperator(ContextTag contextTag, Expr const& expr, Row row, Col col);

    // Add an operator to be assembled on the elements of the grid
    template <class Expr, class Row = RootTreePath, class Col = RootTreePath>
    void addOperator(Expr const& expr, Row row = {}, Col col = {})
    {
      using E = typename RowLocalView::Element;
      addOperator(tag::element_operator<E>{}, expr, row, col);
    }

    // Add an operator to be assembled on the intersections of the grid
    template <class Expr, class Row = RootTreePath, class Col = RootTreePath>
    void addIntersectionOperator(Expr const& expr, Row row = {}, Col col = {})
    {
      using I = typename RowLocalView::GridView::Intersection;
      addOperator(tag::intersection_operator<I>{}, expr, row, col);
    }
    /// @}

    /// Assemble the matrix operators on the bound element.
    void assemble(RowLocalView const& rowLocalView,
                  ColLocalView const& colLocalView);

    /// Assemble all matrix operators, TODO: incorporate boundary conditions
    void assemble(SymmetryStructure symmetry = SymmetryStructure::unknown);

  protected:

    /// Dense matrix to store coefficients during \ref assemble()
    ElementMatrix elementMatrix_;

    /// List of operators associated to row/col node
    MatrixOperators<RowBasis,ColBasis,ElementMatrix> operators_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  template <class RB, class CB>
  BiLinearForm(RB&&, CB&&)
    -> BiLinearForm<Underlying_t<RB>, Underlying_t<CB>>;
#endif

  template <class RB, class CB, class... Args>
  auto makeBiLinearForm(RB&& rowBasis, CB&& colBasis)
  {
    using BLF = BiLinearForm<Underlying_t<RB>, Underlying_t<CB>>;
    return BLF{FWD(rowBasis), FWD(colBasis)};
  }

} // end namespace AMDiS

#include <amdis/BiLinearForm.inc.hpp>
