#pragma once

#include <functional>
#include <list>
#include <memory>
#include <vector>

#include <dune/common/std/optional.hh>

#include <amdis/Boundary.hpp>
#include <amdis/BoundaryCondition.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/typetree/RangeType.hpp>
#include <amdis/typetree/Traversal.hpp>
#include <amdis/typetree/TreeData.hpp>

namespace AMDiS
{
  /// Implements a boundary condition of Dirichlet-type.
  /**
   * By calling the methods \ref init() and \ref finish before and after
   * assembling the system-matrix, respectively, dirichlet boundary conditions
   * can be applied to the matrix and system vector. Therefore, a predicate
   * functions indicates the DOFs where values should be enforced and a second
   * functor provided in the constructor is responsible for determining the
   * values to be set at the DOFs.
   *
   * In the \ref finish method the matrix is called with \ref applyDirichletBC
   * to erase the corresponding rows and columns for the DOF indices. This
   * application of boundary conditions can be symmetric if the matrix does
   * support this symmetric modification. As a result, this method returns a list
   * of columns values, that should be subtracted from the rhs.
   **/
  template <class Domain, class Range = double>
  class DirichletBC
      : public BoundaryCondition
  {
    using Super = BoundaryCondition;

  public:
    template <class BM, class Values,
      REQUIRES(Concepts::Functor<Values, Range(Domain)>) >
    DirichletBC(BM&& boundaryManager, BoundaryType id, Values&& values)
      : Super(FWD(boundaryManager), id)
      , values_(FWD(values))
    {}

    template <class Predicate, class Values,
      REQUIRES(Concepts::Functor<Predicate, bool(Domain)>),
      REQUIRES(Concepts::Functor<Values, Range(Domain)>)>
    DirichletBC(Predicate&& predicate, Values&& values)
      : predicate_(FWD(predicate))
      , values_(FWD(values))
    {}

    template <class IS>
    bool onBoundary(IS const& intersection) const
    {
      return predicate_ ? (*predicate_)(intersection.geometry().center())
                        : Super::onBoundary(intersection);
    }

    // fill \ref dirichletNodes_ with 1 or 0 whether DOF corresponds to the boundary or not.
    template <class RB, class CB>
    void init(RB const& rowBasis, CB const& colBasis);

    /// \brief Apply dirichlet BC to matrix and vector, i.e., add a unit-row
    /// to the matrix and optionally delete the corresponding matrix-column.
    /**
     * \tparam Mat  Matrix
     * \tparam Sol  Vector type of solution
     * \tparam Rhs  Vector type of rhs
     * \tparam RN   RowNode
     * \tparam CN   ColNode
     **/
    template <class Mat, class Sol, class Rhs, class RN, class RTP, class CN, class CTP>
    void fillBoundaryCondition(Mat& matrix, Sol& solution, Rhs& rhs, RN const& rowNode, RTP rowTreePath, CN const& colNode, CTP colTreePath);

  private:
    Dune::Std::optional<std::function<bool(Domain)>> predicate_;
    std::function<Range(Domain)> values_;
    std::vector<bool> dirichletNodes_;
  };


  template <class Basis>
  struct DirichletData
  {
    using WorldVector = typename Basis::GridView::template Codim<0>::Geometry::GlobalCoordinate;

    template <class RN, class CN>
    using type = std::list< std::shared_ptr<DirichletBC<WorldVector, RangeType_t<RN>>> >;
  };

  template <class RowBasis, class ColBasis>
  using DirichletBCs = MatrixData<RowBasis, ColBasis, DirichletData<RowBasis>::template type>;

} // end namespace AMDiS

#include "DirichletBC.inc.hpp"
