#pragma once

#include <algorithm>
#include <list>
#include <memory>
#include <type_traits>

#include <amdis/common/ConceptsBase.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  namespace Impl
  {
    // Forward declaration
    template <class Event>
    class SignalInterface;

    template <class Event>
    class ObserverInterface
    {
    public:
      virtual ~ObserverInterface() = default;

      /// Attach the observer to a subject. It will then receive notifications from the subject when
      /// an event of type Event it triggered.
      void attach(std::shared_ptr<SignalInterface<Event>> const& subject)
      {
        subject->attachObserver(this);
      }

      /// Detach the observer from the subject. It will then no longer receive notifications. This
      /// must be called before the observer is deleted.
      void detach(std::shared_ptr<SignalInterface<Event>> const& subject)
      {
        subject->detachObserver(this);
      }

      /// This method will be called by the subject when triggering the event.
      virtual void update(Event const&)
      {
        error_exit("Method must be overridden by derived class");
      }
    };

    template <class Event>
    class SignalInterface
    {
    public:
      virtual ~SignalInterface() = default;

      /// Attaches an observer to this class. This method will be called by all observers with
      /// with themselves as argument.
      virtual void attachObserver(ObserverInterface<Event>*) = 0;

      /// Detaches an observer to this class. This method will be called by all observers with
      /// with themselves as argument.
      virtual void detachObserver(ObserverInterface<Event>*) = 0;

      /// Notify all observers that have called attachObserver but have not called detachObserver
      virtual void notify(Event const&) const = 0;
    };

    template <class Event>
    class SignalsImpl
        : virtual public SignalInterface<Event>
    {
    public:
      virtual ~SignalsImpl() = default;

      void attachObserver(ObserverInterface<Event>* o) override
      {
        observers_.push_back(o);
      }

      void detachObserver(ObserverInterface<Event>* o) override
      {
        auto it = std::find(observers_.begin(), observers_.end(), o);
        if (it != observers_.end())
          observers_.erase(it);
      }

      void notify(Event const& e) const override
      {
        for (ObserverInterface<Event>* o : observers_)
          o->update(e);
      }

    private:
      std::list<ObserverInterface<Event>*> observers_;
    };

    template <class Event>
    class SignalsDummy
        : virtual public SignalInterface<Event>
    {
    public:
      virtual ~SignalsDummy() = default;
      void attachObserver(ObserverInterface<Event>* o) override {}
      void detachObserver(ObserverInterface<Event>* o) override {}
      void notify(Event const& e) const override {}
    };

    template <class Event>
    class ObserverImpl
        : virtual public ObserverInterface<Event>
    {
      using Self = ObserverImpl;

      ObserverImpl()
        : subject_(std::make_shared<SignalsDummy<Event>>())
      {}

    public:
      ObserverImpl(std::shared_ptr<SignalInterface<Event>> subject)
        : subject_(std::move(subject))
      {
        this->attach(subject_);
      }

      // Get next subject in hierarchy if Event is not handled
      template <class Subject,
        REQUIRES(!std::is_convertible<std::shared_ptr<Subject>, std::shared_ptr<SignalInterface<Event>>>::value)>
      ObserverImpl(std::shared_ptr<Subject> subject)
        : Self(subject->getSubject())
      {}

      /// Copy constructor
      ObserverImpl(Self const& that)
        : subject_(that.subject_)
      {
        this->attach(subject_);
      }

      /// Move constructor
      ObserverImpl(Self&& that)
        : ObserverImpl()
      {
        swap(*this, that);
      }

      /// Destructor. Detaches observer from the subject
      virtual ~ObserverImpl()
      {
        this->detach(subject_);
      }

      /// Copy/Move assignment
      Self& operator=(Self that)
      {
        swap(*this, that);
        return *this;
      }

      /// Swaps the contents of lhs and rhs
      friend void swap(Self& lhs, Self& rhs)
      {
        using std::swap;
        lhs.detach(lhs.subject_);
        rhs.detach(rhs.subject_);
        swap(lhs.subject_, rhs.subject_);
        lhs.attach(lhs.subject_);
        rhs.attach(rhs.subject_);
      }

    private:
      /// The observed subject
      std::shared_ptr<SignalInterface<Event>> subject_;
    };
  }

  /** Mixin for signaling of certain events.
   *  Derived classes T can emit a signal e by calling notify(e). This will send the signal to all
   *  classes U... deriving from Observer<S, Events...> if
   *  - the type of the event is included in Events,
   *  - T = S or S is an Observer of T (directly or indirectly via other Observers),
   *  - U called the Observer constructor with an instance of S that has direct or indirect access
   *      to the instance of T (see Observer)
   */
  template <class Event, class... Events>
  class Signals
      : public Impl::SignalsImpl<Event>
      , public Signals<Events...>
  {
  public:
    using Impl::SignalsImpl<Event>::notify;
    using Signals<Events...>::notify;
  };

  template <class Event>
  class Signals<Event>
      : public Impl::SignalsImpl<Event>
  {
  public:
    using Impl::SignalsImpl<Event>::notify;
  };


  /** Mixin for reacting to certain events
   *  Derived classes T can react to events by implementing the functions update(Event& e) for
   *  all Events specified in the template parameter list, whose origin is an instance of class
   *  Subject specified in the constructor to Observer and any indirectly observed class instances.
   *  Observed instances may include:
   *  - the instance of class Subject passed to the constructor
   *  - the instance passed to the constructor of Subject::Observer<S> if Subject inherits from
   *      Observer<S>
   *  - all other instances indirectly accesible in the above way
   *  For each event E the first object in the above hierarchy that implements Signals<Es> with E
   *  included in Es will be observed by this class.
   */
  template <class Subject, class... Events>
  class Observer
      : public Impl::ObserverImpl<Events>...
  {
    template <class E>
    friend class Impl::ObserverImpl;

  public:
    Observer(std::shared_ptr<Subject> s)
      : Impl::ObserverImpl<Events>(s)...
      , subject_(std::move(s))
    {}

  private:
    /// Get the direct subject of this instance. Used by the implementation to move upwards in the
    /// observer hierarchy.
    std::shared_ptr<Subject> const& getSubject() { return subject_; }

    std::shared_ptr<Subject> subject_;
  };

} // end namespace AMDiS
