#pragma once

#include <list>
#include <memory>

#include <amdis/BoundaryCondition.hpp>
#include <amdis/AssemblerInterface.hpp>
#include <amdis/typetree/TreeData.hpp>

namespace AMDiS
{
  namespace tag
  {
    template <class E> struct element_operator { using type = E; };
    template <class I> struct intersection_operator { using type = I; };
    template <class I> struct boundary_operator : public BoundaryCondition
    {
      using type = I;
      using BoundaryCondition::BoundaryCondition;
    };
  }

  template <class GridView, class Container>
  class OperatorLists
  {
    using Element = typename GridView::template Codim<0>::Entity;
    using Intersection = typename GridView::Intersection;

    template <class Op>
    struct DataElement
    {
      std::shared_ptr<Op> op;
      BoundaryCondition bc;
    };

    /// Lists of \ref DataElement on the Element, BoundaryIntersction, and InteriorIntersections
    template <class... Nodes>
    struct Data
    {
      /// Return whether any operators are stored on the node
      bool empty() const
      {
        return element_.empty() && boundary_.empty() && intersection_.empty();
      }

      /// Test whether to assemble on the node
      bool doAssemble() const
      {
        return !assembled_ || changing_;
      }

      operator bool() const
      {
        return doAssemble() && !empty();
      }

      /// Bind all operators to the grid element and geometry
      template <class Geo>
      void bind(Element const& elem, Geo const& geo)
      {
        for (auto& op : element_) op->bind(elem,geo);
        for (auto& op : intersection_) op->bind(elem,geo);
        for (auto& data : boundary_) data.op->bind(elem,geo);
      }

      /// Unbind all operators from the element
      void unbind()
      {
        for (auto& op : element_) op->unbind();
        for (auto& op : intersection_) op->unbind();
        for (auto& data : boundary_) data.op->unbind();
        assembled_ = true;
      }


      template <class Op>
      void push(tag::element_operator<Element>, Op&& op)
      {
        element_.emplace_back(FWD(op));
      }

      template <class Op>
      void push(tag::intersection_operator<Intersection>, Op&& op)
      {
        intersection_.emplace_back(FWD(op));
      }

      template <class Op>
      void push(tag::boundary_operator<Intersection> b, Op&& op)
      {
        boundary_.push_back({FWD(op), b});
      }


      auto&       onElement()       { return element_; }
      auto const& onElement() const { return element_; }

      auto&       onIntersection()       { return intersection_; }
      auto const& onIntersection() const { return intersection_; }

      auto&       onBoundary()       { return boundary_; }
      auto const& onBoundary() const { return boundary_; }

    private:
      using ElementTraits = DefaultAssemblerTraits<Element,Container>;
      using IntersectionTraits = DefaultAssemblerTraits<Intersection,Container>;

      /// The type of local operators associated with grid elements
      using ElementAssembler = AssemblerInterface<ElementTraits, Nodes...>;

      /// The type of local operators associated with grid intersections
      using IntersectionAssembler = AssemblerInterface<IntersectionTraits, Nodes...>;

      /// List of operators to be assembled on grid elements
      std::vector<std::shared_ptr<ElementAssembler>> element_;

      /// List of operators to be assembled on interior intersections
      std::vector<std::shared_ptr<IntersectionAssembler>> intersection_;

      /// List of operators to be assembled on boundary intersections
      std::vector<DataElement<IntersectionAssembler>> boundary_;

      /// if false, do reassemble of all operators
      bool assembled_ = false;

      /// if true, or \ref assembled false, do reassemble of all operators
      bool changing_ = true;
    };

  public:

    /// List of operators associated with matrix blocks (RowNode, ColNode)
    template <class RowNode, class ColNode>
    using MatData = Data<RowNode, ColNode>;

    /// List of operators associated with vector blocks [Node]
    template <class Node>
    using VecData = Data<Node>;
  };


  template <class RowBasis, class ColBasis, class ElementMatrix>
  using MatrixOperators
    = MatrixData<RowBasis, ColBasis, OperatorLists<typename RowBasis::GridView,ElementMatrix>::template MatData>;

  template <class GlobalBasis, class ElementVector>
  using VectorOperators
    = VectorData<GlobalBasis, OperatorLists<typename GlobalBasis::GridView,ElementVector>::template VecData>;

} // end namespace AMDiS
