#pragma once

#include <dune/functions/functionspacebases/basistags.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/grid/yaspgrid.hh>

#include <amdis/common/Logical.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/functions/ParallelGlobalBasis.hpp>

namespace AMDiS
{
  namespace Impl
  {
    template <bool same, int... degs>
    struct LagrangePreBasisCreatorImpl;

    // specialization for composite basis
    template <int... degs>
    struct LagrangePreBasisCreatorImpl<false, degs...>
    {
      static auto create()
      {
        using namespace Dune::Functions::BasisFactory;
        return composite(lagrange<degs>()..., flatLexicographic());
      }
    };

    // specialization for power basis
    template <int deg, int... degs>
    struct LagrangePreBasisCreatorImpl<true, deg, degs...>
    {
      static auto create()
      {
        using namespace Dune::Functions::BasisFactory;
        return power<1+sizeof...(degs)>(lagrange<deg>(), flatLexicographic());
      }
    };

    // factory to construct a global basis of several lagrange bases, with flat indexing.
    template <int deg, int... degs>
    struct LagrangePreBasisCreator
        : public LagrangePreBasisCreatorImpl<all_of_v<(deg == degs)...>, deg, degs...>
    {};


    template <int dow, int k = 1>
    struct TaylorHoodPreBasisCreator
    {
      static auto create()
      {
        using namespace Dune::Functions::BasisFactory;
        return composite(power<dow>(lagrange<k+1>(), flatInterleaved()), lagrange<k>(), flatLexicographic());
      }
    };

  } // end namespace Impl

  /// An Exemplar for ProblemStatTraits
  template <class GlobalBasisType, class T = double>
  struct DefaultProblemTraits
  {
    using GlobalBasis = GlobalBasisType;
    using CoefficientType = T;
  };


  template <class Grid, class PreBasisCreator, class T = double>
  struct DefaultBasisCreator
  {
    using GridView = typename Grid::LeafGridView;
    static auto create(std::string const& name, GridView const& gridView)
    {
      return makeGlobalBasis(name, gridView, PreBasisCreator::create());
    }

    static auto create(GridView const& gridView)
    {
      return makeGlobalBasis(gridView, PreBasisCreator::create());
    }

    using GlobalBasis = Underlying_t<decltype(create(std::declval<GridView>()))>;
    using CoefficientType = T;
  };

  /// \brief ProblemStatTraits for a (composite) basis composed of
  /// lagrange bases of different degree.
  template <class Grid, int... degrees>
  struct LagrangeBasis
      : public DefaultBasisCreator<Grid, Impl::LagrangePreBasisCreator<degrees...>>
  {};

  /// \brief Specialization of \ref LagrangeBasis for Grid type
  /// \ref Dune::YaspGrid for a given dimension.
  template <int dim, int... degrees>
  struct YaspGridBasis
      : public LagrangeBasis<Dune::YaspGrid<dim>, degrees...>
  {};

  /// \brief ProblemStatTraits of Taylor-Hood basis of lagrange-type
  /// with pressure degree k
  template <class Grid, int k = 1>
  struct TaylorHoodBasis
      : public DefaultBasisCreator<Grid, Impl::TaylorHoodPreBasisCreator<Grid::dimensionworld,k>>
  {};

} // end namespace AMDiS
