#pragma once

#include <dune/common/typeutilities.hh>

#include <amdis/common/Concepts.hpp>
#include <amdis/common/Logical.hpp>
#include <amdis/common/TypeTraits.hpp>

namespace AMDiS
{
  namespace Concepts
  {
    namespace Definition
    {
      struct HasVectorAccess
      {
        template <class V, class I>
        auto requires_(V&& v, I&& i, Dune::PriorityTag<2>) -> decltype( v[i] );

        template <class V, class I>
        auto requires_(V&& v, I&& i, Dune::PriorityTag<1>) -> decltype( v(i) );
      };

      struct HasMatrixAccess
      {
        template <class M, class I, class J>
        auto requires_(M&& m, I&& i, J&& j, Dune::PriorityTag<2>) -> decltype( m[i][j] );

        template <class M, class I, class J>
        auto requires_(M&& m, I&& i, J&& j, Dune::PriorityTag<1>) -> decltype( m(i,j) );
      };

    } // end namespace Definition

    /// Vector component can be accessed either with [.] or (.)
    template <class V, class I>
    using VectorAccessible_t = bool_t<models<Definition::HasVectorAccess(V, I, Dune::PriorityTag<42>)>>;

    /// Matrix component can be accessed either with [.][.] or (.,.)
    template <class M, class I, class J>
    using MatrixAccessible_t = bool_t<models<Definition::HasMatrixAccess(M, I, J, Dune::PriorityTag<42>)>>;

  } // end namespace Concepts


#ifdef DOXYGEN
  /// \brief Uniform vector access using [.] or (.)
  template <class Vector, class I>
  decltype(auto) access(Vector&& vec, I const& i);

  /// \brief Uniform matrix access using either [.][.] or (.,.)
  template <class Matrix, class I, class J>
  decltype(auto) access(Matrix&& mat, I const& i, J const& j);
#else

  namespace Impl
  {
    // access i'th component of a vector using [.]
    template <class Vector, class I,
      class = void_t<decltype(std::declval<Vector>()[std::declval<I>()])> >
    decltype(auto) access_impl(Vector&& vec, I const& i, Dune::PriorityTag<2>)
    {
      return vec[i];
    }

    // access i'th component of a vector using (.)
    template <class Vector, class I,
      class = void_t<decltype(std::declval<Vector>()(std::declval<I>()))> >
    decltype(auto) access_impl(Vector&& vec, I const& i, Dune::PriorityTag<1>)
    {
      return vec(i);
    }

    // fall-back implementation for scalars
    template <class Vector, class I>
    decltype(auto) access_impl(Vector&& vec, I const& /*i*/, Dune::PriorityTag<0>)
    {
      return FWD(vec);
    }
  }

  // access i'th component of a vector
  template <class Vector, class I>
  decltype(auto) access(Vector&& vec, I const& i)
  {
    return Impl::access_impl(FWD(vec), i, Dune::PriorityTag<42>{});
  }


  namespace Impl
  {
    // access (i,j)'th component of a matrix using [.][.]
    template <class Matrix, class I, class J,
      class = void_t<decltype(std::declval<Matrix>()[std::declval<I>(),std::declval<J>()])> >
    decltype(auto) access_impl(Matrix&& mat, I const& i, J const& j, Dune::PriorityTag<2>)
    {
      return mat[i][j];
    }

    // access (i,j)'th component of a matrix using (.,.)
    template <class Matrix, class I, class J,
      class = void_t<decltype(std::declval<Matrix>()(std::declval<I>(),std::declval<J>()))> >
    decltype(auto) access_impl(Matrix&& mat, I const& i, J const& j, Dune::PriorityTag<1>)
    {
      return mat(i,j);
    }

    // fall-back implementation for scalars
    template <class Matrix, class I, class J>
    decltype(auto) access_impl(Matrix&& mat, I const& /*i*/, J const& /*j*/, Dune::PriorityTag<0>)
    {
      return FWD(mat);
    }
  }

  // access (i,j)'th component of a matrix
  template <class Matrix, class I, class J>
  decltype(auto) access(Matrix&& mat, I const& i, J const& j)
  {
    return Impl::access_impl(FWD(mat), i, j, Dune::PriorityTag<42>{});
  }

#endif

} // end namespace AMDiS
