#pragma once

#include <algorithm>
#include <cmath>
#include <cstdint>
#include <limits>
#include <type_traits>

#include <dune/common/power.hh>

namespace AMDiS
{
  namespace Math
  {
    /// Implementation of the absolute value \f$|a|\f$ of arithmetic types.
    template <class T>
    constexpr T abs(T const& a)
    {
      return  a < 0 ? -a : a;
    }


    /// Implementation of the square \f$ a^2 \f$ of arithmetic types.
    template <class T>
    constexpr auto sqr(T const& a)
    {
      return a*a;
    }

    /// Implementation of the power \f$ v^p \f$ of arithmetic types `T`.
    template <int p, class T>
    constexpr auto pow(T const& v)
    {
      static_assert( p >= 0, "Exponent p in `pow<p>(v)` should be >= 0," );
      return Dune::Power<p>::eval(v);
    }


    /// Implementation of the minimum of values \f$ min(a,b)\f$ of any type
    /// supporting the `<` relation.
    /// @{

    template <class T0, class T1>
    constexpr auto min(T0 a, T1 b)
    {
      using T = std::conditional_t<(sizeof(T0)>sizeof(T1)),T0,T1>;
      return a < b ? T(a) : T(b);
    }

    template <class T0>
    constexpr T0 min(T0 a)
    {
      return a;
    }

    template <class T0, class... Ts>
    constexpr auto min(T0 a, Ts... ts)
    {
      return min(a, min(ts...));
    }

    /// @}


    /// Implementation of the maximum of values \f$ max(a,b)\f$ of any type
    /// supporting the `<` relation.
    /// @{

    template <class T0, class T1>
    constexpr auto max(T0 a, T1 b)
    {
      using T = std::conditional_t<(sizeof(T0)>sizeof(T1)),T0,T1>;
      return a < b ? T(b) : T(a);
    }

    template <class T0>
    constexpr T0 max(T0 a)
    {
      return a;
    }

    template <class T0, class... Ts>
    constexpr auto max(T0 a, Ts... ts)
    {
      return max(a, max(ts...));
    }

    /// @}

    // sum over empty set is zero
    constexpr double sum()
    {
      return 0.0;
    }

#ifndef AMDIS_HAS_CXX_FOLD_EXPRESSIONS
    template <class T>
    constexpr auto const& sum(T const& t)
    {
      return t;
    }

    template <class T0, class T1>
    constexpr auto sum(T0 const& t0, T1 const& t1)
    {
      return t0 + t1;
    }

    template <class T0, class... Ts>
    constexpr auto sum(T0 const& t0, Ts const&... ts)
    {
      return t0 + sum(ts...);
    }
#else
    template <class... Ts>
    constexpr auto sum(Ts const&... ts)
    {
      return (ts + ...);
    }
#endif

  } // end namespace Math

  template <class T>
  constexpr T threshold = std::numeric_limits<T>::epsilon();

} // end namespace AMDiS
