#pragma once

#include <array>
#include <tuple>
#include <type_traits>

#include <amdis/common/TypeTraits.hpp>

namespace Dune
{
  // forward declarations
  template <class T, int N, int M>
  class FieldMatrix;

  template <class T, int N>
  class FieldVector;

  template <class... Ts>
  class TupleVector;

  template <class Row0, class... Rows>
  class MultiTypeBlockMatrix;

  template <class... Ts>
  class MultiTypeBlockVector;
}

namespace Eigen
{
  template <typename _Scalar, int _Rows, int _Cols, int _Options, int _MaxRows, int _MaxCols>
  class Matrix;
}

namespace AMDiS
{
  namespace Impl
  {
    template <class Tuple, class = void>
    struct SizeImpl : std::integral_constant<std::size_t, 0> {};

    template <class... Args>
    struct SizeImpl<std::tuple<Args...>>
        : std::integral_constant<std::size_t, sizeof...(Args) > {};

    template <class Arg0, class Arg1>
    struct SizeImpl<std::pair<Arg0, Arg1>>
        : std::integral_constant<std::size_t, 2> {};

    template <class T, std::size_t N>
    struct SizeImpl<std::array<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N>
    struct SizeImpl<Dune::FieldVector<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N, int M>
    struct SizeImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, N*M> {};

    template <class... Ts>
    struct SizeImpl<Dune::TupleVector<Ts...>>
        : std::integral_constant<std::size_t, sizeof...(Ts)> {};

    template <class... Ts>
    struct SizeImpl<Dune::MultiTypeBlockVector<Ts...>>
        : std::integral_constant<std::size_t, sizeof...(Ts)> {};

    template <class T, int N, int... opts>
    struct SizeImpl<Eigen::Matrix<T,N,1,opts...>>
        : std::integral_constant<std::size_t, (N >= 0 ? std::size_t(N) : 0u)> {};

    template <class T, int N, int... opts>
    struct SizeImpl<Eigen::Matrix<T,1,N,opts...>>
        : std::integral_constant<std::size_t, (N >= 0 ? std::size_t(N) : 0u)> {};

    // Specialization for arithmetic types
    template <class T>
    struct SizeImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

  } // end namespace Impl

  /// Get the number of elements in a tuple / pair / array / ...
  template <class T>
  constexpr std::size_t Size_v = Impl::SizeImpl<remove_cvref_t<T>>::value;

  template <class T>
  using Size_t = Impl::SizeImpl<remove_cvref_t<T>>;


  namespace Impl
  {
    template <class Tuple, class = void>
    struct RowsImpl : std::integral_constant<std::size_t, 0> {};

    template <class T, int N, int M>
    struct RowsImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, N> {};

    template <class... Rows>
    struct RowsImpl<Dune::MultiTypeBlockMatrix<Rows...>>
        : std::integral_constant<std::size_t, sizeof...(Rows)> {};

    // Specialization for arithmetic types
    template <class T>
    struct RowsImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

    template <class T, int N, int M, int... opts>
    struct RowsImpl<Eigen::Matrix<T,N,M,opts...>>
        : std::integral_constant<std::size_t, (N >= 0 ? std::size_t(N) : 0u)> {};

  } // end namespace Impl

  /// Get the number of rows in a fixed-size matrix
  template <class T>
  constexpr std::size_t Rows_v = Impl::RowsImpl<remove_cvref_t<T>>::value;

  template <class T>
  using Rows_t = Impl::RowsImpl<remove_cvref_t<T>>;


  namespace Impl
  {
    template <class Tuple, class = void>
    struct ColsImpl : std::integral_constant<std::size_t, 0> {};

    template <class T, int N, int M>
    struct ColsImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, M> {};

    template <class Row0, class... Rows>
    struct ColsImpl<Dune::MultiTypeBlockMatrix<Row0, Rows...>>
        : SizeImpl<Row0> {};

    // Specialization for arithmetic types
    template <class T>
    struct ColsImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

    template <class T, int N, int M, int... opts>
    struct ColsImpl<Eigen::Matrix<T,N,M,opts...>>
        : std::integral_constant<std::size_t, (M >= 0 ? std::size_t(M) : 0u)> {};

  } // end namespace Impl

  /// Get the number of columns in a fixed-size matrix
  template <class T>
  constexpr std::size_t Cols_v = Impl::ColsImpl<remove_cvref_t<T>>::value;

  template <class T>
  using Cols_t = Impl::ColsImpl<remove_cvref_t<T>>;

} // end namespace AMDiS
