#pragma once


#include <algorithm>
#include <list>
#include <memory>
#include <type_traits>
#include <utility>

#include <dune/common/concept.hh>
#include <dune/common/reservedvector.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/common/typeutilities.hh>
#include <dune/common/version.hh>

#include <dune/functions/common/type_traits.hh>
#include <dune/functions/functionspacebases/concepts.hh>
#include <dune/functions/functionspacebases/defaultlocalview.hh>
#include <dune/functions/functionspacebases/flatmultiindex.hh>

#include <dune/grid/common/adaptcallback.hh>

#include <dune/typetree/treepath.hh>

#include <amdis/common/Concepts.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/linearalgebra/Traits.hpp>
#include <amdis/AdaptiveGrid.hpp>
#include <amdis/Observer.hpp>
#include <amdis/Output.hpp>

#if DUNE_VERSION_LT(DUNE_FUNCTIONS,2,7)
#include <dune/functions/functionspacebases/defaultlocalindexset.hh>
#endif

namespace Dune
{
  namespace Functions
  {
    // Forward declaration
    template <class PB>
    class DefaultGlobalBasis;
  }
}

namespace AMDiS
{
  namespace event
  {
    template <class B>
    struct basisUpdate
    {
      basisUpdate(B const& b_)
        : basis(b_)
      {}

      B const& basis;
    };
  }

  /**
   * \brief Parallel global basis defined on a (sequential) pre-basis
   *
   * This class is an expansion to Dune::Functions::DefaultGlobalBasis<PB>. It adds a communicator
   * to use the basis in parallel as well as automatic update functionality.
   *
   * \tparam PB  Pre-basis providing the implementation details
   */
  template <class PB>
  class ParallelGlobalBasis
        : public Observer<AdaptiveGrid<typename PB::GridView::Grid>, event::adapt>
        , public Signals<event::basisUpdate<ParallelGlobalBasis<PB>>>
  {
    using Subject = AdaptiveGrid<typename PB::GridView::Grid>;
    using Obs = Observer<Subject, event::adapt>;
    using Self = ParallelGlobalBasis<PB>;

    struct DummyImpl {};

  public:

    /// Pre-basis providing the implementation details
    using PreBasis = PB;

    /// The empty prefix path that identifies the root in the local ansatz tree
    using PrefixPath = Dune::TypeTree::HybridTreePath<>;

    /// The grid view that the FE space is defined on
    using GridView = typename PreBasis::GridView;
    // using Grid = typename GridView::Grid; // TODO(FM): Use once AdaptiveGrid is complete Metagrid
    using Grid = Subject;

    /// Type used for global numbering of the basis vectors
    using MultiIndex = typename PreBasis::MultiIndex;

    /// Type used for indices and size information
    using size_type = std::size_t;

    /// Type of the local view on the restriction of the basis to a single element
    using LocalView = Dune::Functions::DefaultLocalView<Self>;

    /// Type used for prefixes handed to the size() method
    using SizePrefix = typename PreBasis::SizePrefix;

#if DUNE_VERSION_LT(DUNE_FUNCTIONS,2,7)
    /// Node index set provided by PreBasis
    using NodeIndexSet = typename PreBasis::template IndexSet<PrefixPath>;

    /// Type of local indixes set exported by localIndexSet()
    using LocalIndexSet = Dune::Functions::DefaultLocalIndexSet<LocalView, NodeIndexSet>;
#endif

    /// Type of the communicator
    using Comm = typename BackendTraits<Self>::Comm;

    using ADH = Dune::AdaptDataHandle<Grid, DummyImpl>;

    /**
     * \brief Constructor
     *
     * \tparam T Argument list for PreBasis
     * \param s The AdaptiveGrid providing the GridView for this basis
     * \param t Argument list for PreBasis
     *
     * This will forward all arguments to the constructor of PreBasis
     */
    template<class... T,
      Dune::Functions::enableIfConstructible<PreBasis, T...> = 0>
    ParallelGlobalBasis(std::string const& name, std::shared_ptr<Subject> s, T&&... t)
      : Obs(std::move(s))
      , preBasis_(FWD(t)...)
      , prefixPath_()
    {
      static_assert(Dune::models<Dune::Functions::Concept::PreBasis<GridView>, PreBasis>(),
        "Type passed to ParallelGlobalBasis does not model the PreBasis concept.");
      preBasis_.initializeIndices();
      comm_ = CommunicationCreator<Comm>::create(*this, name + "->solver");
    }

    template <class... T>
    ParallelGlobalBasis(std::string const& name, Subject& s, T&&... t)
      : ParallelGlobalBasis(name, Dune::stackobject_to_shared_ptr(s), FWD(t)...)
    {}

    template <class S, class... T,
    REQUIRES(Concepts::Same<Underlying_t<S>, Subject>)>
    ParallelGlobalBasis(S&& s, T&&... t)
      : ParallelGlobalBasis("", FWD(s), FWD(t)...)
    {}

    /** Converting constructor from Dune::DefaultGlobalBasis<PB>.
     *  This will create a new ParallellGlobalBasis. The pre-basis is copied from the constructor
     *  argument and a new communication object is built.
     */
    // TODO(FM): Replace explicit type with concept
    ParallelGlobalBasis(Dune::Functions::DefaultGlobalBasis<PB>&& from)
      : Obs(Subject::instance(from.gridView().grid()))
      , preBasis_(from.preBasis())
      , prefixPath_(from.prefixPath())
    {
      comm_ = CommunicationCreator<Comm>::create(*this);
    }

    /// Obtain the grid view that the basis is defined on
    GridView const& gridView() const
    {
      return preBasis_.gridView();
    }

    /// Obtain the pre-basis providing the implementation details
    PreBasis const& preBasis() const
    {
      return preBasis_;
    }

    /// Obtain the pre-basis providing the implementation details
    PreBasis& preBasis()
    {
      return preBasis_;
    }

    /// Updates the underlying basis when event::adapt is triggered by the observed grid
    void update(event::adapt const& e) override
    {
      if (e.adapted)
      {
        update(gridView());
        event::basisUpdate<Self> eNew{*this};
        this->notify(eNew);
      }
    }

    /**
     * \brief Update the stored grid view
     *
     * This will update the indexing information of the global basis as well as the communicator.
     * It is called automatically if the grid has changed.
     */
    void update(GridView const& gv)
    {
      preBasis_.update(gv);
      preBasis_.initializeIndices();
      comm_->update(*this);
    }

    /// Get the total dimension of the space spanned by this basis
    size_type dimension() const
    {
      return preBasis_.dimension();
    }

    /// Return number of possible values for next position in empty multi index
    size_type size() const
    {
      return preBasis_.size();
    }

    /// Return number of possible values for next position in multi index
    size_type size(const SizePrefix& prefix) const
    {
      return preBasis_.size(prefix);
    }

    /// Return local view for basis
    LocalView localView() const
    {
      return LocalView(*this);
    }

    /// Return *this because we are not embedded in a larger basis
    ParallelGlobalBasis const& rootBasis() const
    {
      return *this;
    }

    /// Return empty path, because this is the root in the local ansatz tree
    PrefixPath const& prefixPath() const
    {
      return prefixPath_;
    }

    /// Return the communicator.
    /// This provides the means to communicate data associated to the basis with other processes.
    Comm const& communicator() const  { return *comm_; }
    Comm&       communicator()        { return *comm_; }

    ADH& globalRefineCallback() const
    {
      // TODO(FM): Implement
      error_exit("Not implemented: ParallelGlobalBasis::globalRefineCallback()");
      return ADH{};
    }

  protected:
    PreBasis preBasis_;
    PrefixPath prefixPath_;
    std::unique_ptr<Comm> comm_;
  };

  template<class MultiIndex, class GridView, class PreBasisFactory>
  auto makeGlobalBasis(std::string const& name, GridView const& gridView,
                       PreBasisFactory&& preBasisFactory)
  {
    auto preBasis = preBasisFactory.template makePreBasis<MultiIndex>(gridView);
    using PreBasis = std::decay_t<decltype(preBasis)>;

    using Grid = typename GridView::Grid;
    auto ag = AdaptiveGrid<Grid>::instance(gridView.grid());

    return ParallelGlobalBasis<PreBasis>(name, std::move(ag), std::move(preBasis));
  }

  template<class GridView, class PreBasisFactory>
  auto makeGlobalBasis(std::string const& name, GridView const& gridView,
                       PreBasisFactory&& preBasisFactory)
  {
    using RawPreBasisFactory = std::decay_t<PreBasisFactory>;
    using MultiIndex = std::conditional_t<
      (RawPreBasisFactory::requiredMultiIndexSize == 1),
      Dune::Functions::FlatMultiIndex<std::size_t>,
      Dune::ReservedVector<std::size_t, RawPreBasisFactory::requiredMultiIndexSize>>;

    return makeGlobalBasis<MultiIndex, GridView, PreBasisFactory>(name, gridView,
      FWD(preBasisFactory));
  }

  template<class GridView, class PreBasisFactory>
  auto makeGlobalBasis(GridView const& gridView, PreBasisFactory&& preBasisFactory)
  {
    return makeGlobalBasis("", gridView, FWD(preBasisFactory));
  }

} // end namespace AMDiS
