#pragma once

#include <vector>

#include <dune/common/std/optional.hh>
#include <dune/functions/common/defaultderivativetraits.hh>
#include <dune/functions/functionspacebases/defaultnodetorangemap.hh>
#include <dune/functions/functionspacebases/flatvectorview.hh>
#include <dune/functions/gridfunctions/gridviewentityset.hh>
#include <dune/typetree/childextraction.hh>

#include <amdis/LinearAlgebra.hpp>
#include <amdis/typetree/FiniteElementType.hpp>
#include <amdis/typetree/RangeType.hpp>
#include <amdis/typetree/TreePath.hpp>

namespace AMDiS
{
  /// \class DiscreteFunction
  /// \brief A view on a subspace of a \ref DOFVector
  /**
   * \ingroup GridFunctions
   *
   * \tparam GB  Type of the global basis
   * \tparam VT  Coefficient type of the DOFVector
   * \tparam TP  A realization of \ref Dune::TypeTree::HybridTreePath
   *
   * **Requirements:**
   * - GB models \ref Dune::Functions::Concept::GlobalBasis
   **/
  template <class GB, class VT, class TP>
  class DiscreteFunction
  {
    // static_assert(Dune::IsNumber<VT>::value, "");

  private:
    using GlobalBasis = GB;
    using TreePath = TP;

    using Tree = typename GlobalBasis::LocalView::Tree;
    using SubTree = typename Dune::TypeTree::ChildForTreePath<Tree, TreePath>;
    using NodeToRangeEntry = Dune::Functions::DefaultNodeToRangeMap<SubTree>;

    using GridView = typename GlobalBasis::GridView;

  public:
    /// Set of entities the DiscreteFunction is defined on
    using EntitySet = Dune::Functions::GridViewEntitySet<GridView, 0>;

    /// Global coordinates of the EntitySet
    using Domain = typename EntitySet::GlobalCoordinate;

    /// Range type of this DiscreteFunction
    using Range = RangeType_t<SubTree,VT>;

    /// \brief This GridFunction has no derivative function, it can be created
    /// by \ref DiscreteGridFunction.
    enum { hasDerivative = false };

  public:
    /// A LocalFunction representing the derivative of the DOFVector on a bound element
    template <class Type>
    class DerivativeLocalFunctionBase;

    class GradientLocalFunction;
    class PartialLocalFunction;
    class DivergenceLocalFunction;

    /// A LocalFunction representing the value the DOFVector on a bound element
    class LocalFunction;

  public:
    /// Constructor. Stores a pointer to the dofVector and a copy of the treePath.
    DiscreteFunction(DOFVector<GB,VT> const& dofVector, TP const& treePath)
      : dofVector_(&dofVector)
      , treePath_(treePath)
      , entitySet_(dofVector.basis()->gridView())
      , nodeToRangeEntry_(Dune::Functions::makeDefaultNodeToRangeMap(*dofVector.basis(), treePath))
    {}

    /// Constructor forwards to the treePath constructor, with empty TreePath
    DiscreteFunction(DOFVector<GB,VT> const& dofVector)
      : DiscreteFunction(dofVector, Dune::TypeTree::hybridTreePath())
    {}

    /// \brief Evaluate DiscreteFunction in global coordinates. NOTE: expensive
    Range operator()(Domain const& x) const;

    /// \brief Create a local function for this view on the DOFVector. \relates LocalFunction
    LocalFunction makeLocalFunction() const
    {
      return LocalFunction{*this};
    }

    /// \brief Return a \ref Dune::Functions::GridViewEntitySet
    EntitySet const& entitySet() const
    {
      return entitySet_;
    }

  public:
    /// \brief Return global basis bound to the DOFVector
    std::shared_ptr<GlobalBasis const> basis() const
    {
      return dofVector_->basis();
    }

    /// \brief Return treePath associated with this discrete function
    TreePath const& treePath() const
    {
      return treePath_;
    }

    /// \brief Return const coefficient vector
    DOFVector<GB,VT> const& coefficients() const
    {
      return *dofVector_;
    }

  protected:
    DOFVector<GB,VT> const* dofVector_;
    TreePath treePath_;
    EntitySet entitySet_;
    NodeToRangeEntry nodeToRangeEntry_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  // Deduction guide for DiscreteFunction class
  template <class GlobalBasis, class ValueType>
  DiscreteFunction(DOFVector<GlobalBasis, ValueType> const& dofVector)
    -> DiscreteFunction<GlobalBasis, ValueType, Dune::TypeTree::HybridTreePath<>>;
#endif

  /// A Generator for a \ref DiscreteFunction
  template <class GlobalBasis, class ValueType, class PreTreePath>
  auto makeDiscreteFunction(DOFVector<GlobalBasis, ValueType> const& dofVector, PreTreePath const& preTreePath)
  {
    auto treePath = makeTreePath(preTreePath);
    return DiscreteFunction<GlobalBasis, ValueType, decltype(treePath)>{dofVector, treePath};
  }

  /// A Generator for a \ref DiscreteFunction
  template <class GlobalBasis, class ValueType>
  auto makeDiscreteFunction(DOFVector<GlobalBasis, ValueType> const& dofVector)
  {
    auto treePath = Dune::TypeTree::hybridTreePath();
    return DiscreteFunction<GlobalBasis, ValueType, Dune::TypeTree::HybridTreePath<>>{dofVector, treePath};
  }

} // end namespace AMDiS

#include "DiscreteFunction.inc.hpp"
