#pragma once

#include <memory>
#include <type_traits>

#include <dune/common/shared_ptr.hh>

#include <amdis/common/Concepts.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/functions/NodeIndices.hpp>
#include <amdis/linearalgebra/SymmetryStructure.hpp>
#include <amdis/typetree/MultiIndex.hpp>

namespace AMDiS
{
  /**
   * Basis implementation of DOFMatrix, i.e. a sparse matrix storing all the
   * assembled Operators indexed with DOF indices. The matrix data is associated
   * to a row and column global basis.
   *
   * \tparam RB  Basis of the matrix rows
   * \tparam CB  Basis of matrix columns
   * \tparam B   A linear-algebra backend for the matrix storage
   **/
  template <class RB, class CB, class B>
  class MatrixBase
  {
    using Self = MatrixBase;

  public:
    /// The type of the finite element space / basis of the row
    using RowBasis = RB;
    using RowLocalView = typename RowBasis::LocalView;

    /// The type of the finite element space / basis of the column
    using ColBasis = CB;
    using ColLocalView = typename ColBasis::LocalView;

    /// The Linear-Algebra backend used to store the assembled coefficients
    using Backend = B;

  public:
    /// (1) Constructor. Stores the shared_ptr to the bases and the communication object.
    MatrixBase(std::shared_ptr<RB> rowBasis, std::shared_ptr<CB> colBasis)
      : rowBasis_(std::move(rowBasis))
      , colBasis_(std::move(colBasis))
      , backend_(*rowBasis_, *colBasis_)
    {}

    /// (2) Constructor. Forwards to (1) by wrapping reference into non-destroying
    /// shared_ptr, see \ref Dune::wrap_or_move.
    template <class RB_, class CB_,
      REQUIRES(Concepts::Similar<Types<RB_,CB_>, Types<RB,CB>>)>
    MatrixBase(RB_&& rowBasis, CB_&& colBasis)
      : MatrixBase(Dune::wrap_or_move(FWD(rowBasis)), Dune::wrap_or_move(FWD(colBasis)))
    {}

    /// Return the row-basis \ref rowBasis of the matrix
    std::shared_ptr<RowBasis> const& rowBasis() const
    {
      return rowBasis_;
    }

    /// Return the col-basis \ref colBasis of the matrix
    std::shared_ptr<ColBasis> const& colBasis() const
    {
      return colBasis_;
    }

    /// Return the underlying linear algebra backend
    Backend const& backend() const { return backend_; }
    Backend&       backend()       { return backend_; }

    /// \brief Initialize the matrix for insertion, i.e. allocate the non-zero pattern
    /**
     * With the optional parameter \p symmetry some additional information about the
     * structure of the values or the sparsity pattern can be provided.
     * See \ref SymmetryStructure.
     **/
    void init(SymmetryStructure symmetry = SymmetryStructure::unknown)
    {
      backend_.init(*rowBasis_, *colBasis_, symmetry);
    }

    /// Finish the matrix insertion, e.g. cleanup or final insertion
    void finish()
    {
      backend_.finish();
    }

    /// Insert a single value into the matrix (add to existing value)
    template <class RowIndex, class ColIndex>
    void insert(RowIndex const& row, ColIndex const& col, typename Backend::value_type const& value)
    {
      backend_.insert(row, col, value);
    }

    /// Insert a block of values into the sparse matrix (add to existing values)
    /// The global matrix indices are determined by the corresponding localviews.
    template <class LocalMatrix>
    void scatter(RowLocalView const& r, ColLocalView const& c, LocalMatrix const& localMatrix)
    {
      assert(r.size() * c.size() == localMatrix.size());
      assert(r.size() == localMatrix.rows());
      assert(c.size() == localMatrix.cols());

      const bool optimized = std::is_same<RowLocalView,ColLocalView>::value
        && r.tree().treeIndex() == c.tree().treeIndex();

      if (optimized)
        backend_.scatter(nodeIndices(r), localMatrix);
      else
        backend_.scatter(nodeIndices(r), nodeIndices(c), localMatrix);
    }

    /// Number of nonzeros in the matrix
    std::size_t nnz() const
    {
      return backend_.nnz();
    }

  protected:
    /// The finite element space / basis associated with the rows
    std::shared_ptr<RowBasis> rowBasis_;

    /// The finite element space / basis associated with the columns
    std::shared_ptr<ColBasis> colBasis_;

    /// Data backend
    Backend backend_;
  };

} // end namespace AMDiS
