#pragma once

#include <memory>
#include <string>
#include <type_traits>
#include <utility>

#include <dune/common/classname.hh>
#include <dune/common/deprecated.hh>
#include <dune/common/hybridutilities.hh>
#include <dune/common/rangeutilities.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/common/std/type_traits.hh>
#include <dune/functions/functionspacebases/concepts.hh>
#include <dune/functions/functionspacebases/sizeinfo.hh>

#include <amdis/Output.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/common/FakeContainer.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/functions/NodeIndices.hpp>
#include <amdis/operations/Assigner.hpp>
#include <amdis/typetree/MultiIndex.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base vector and a corresponding basis
  /**
   * A vector storing all the assembled Operators indexed with DOF indices. The
   * vector data is associated to a global basis.
   *
   * \tparam GB  Basis of the vector
   * \tparam B   A linear algebra backend implementing the storage and operations.
   **/
  template <class GB, class B>
  class VectorBase
  {
    using Self = VectorBase;

  public:
    /// The type of the functionspace basis associated to this linearform
    using GlobalBasis = GB;
    using LocalView = typename GlobalBasis::LocalView;

    /// The Linear-Algebra backend used to store the assembled coefficients
    using Backend = B;

  private:
    // proxy class redirecting the mutable element access to the insertValue method
    template <class Index>
    struct AccessProxy;

    enum class VectorState
    {
      unknown = 0,
      synchronized = 1,
      insert_values = 2,
      add_values = 3
    };

    template <class A>
    using VectorStateOf_t = std::conditional_t<std::is_same<A,Assigner::plus_assign>::value,
      std::integral_constant<VectorState, VectorState::add_values>,
      std::integral_constant<VectorState, VectorState::insert_values>>;

  public:
    /// (1) Constructor. Stores the shared_ptr of the basis.
    VectorBase(std::shared_ptr<GB> basis)
      : basis_(std::move(basis))
      , backend_(*basis_)
    {
      resizeZero();
    }

    /// (2) Constructor. Forwards to (1) by wrapping reference into non-destroying shared_ptr.
    template <class GB_,
      REQUIRES(Concepts::Similar<GB_,GB>)>
    VectorBase(GB_&& basis)
      : VectorBase(Dune::wrap_or_move(FWD(basis)))
    {}

    /// Return the basis \ref basis_ associated to the vector
    std::shared_ptr<GlobalBasis> const& basis() const { return basis_; }
    std::shared_ptr<GlobalBasis> const& basis()       { return basis_; }

    /// Return the underlying linear algebra backend
    Backend const& backend() const { return backend_; }
    Backend&       backend()       { return backend_; }


    template <class B_>
    using HasLocalSize = decltype(std::declval<B_>().localSize());

    template <class B_>
    using HasGlobalSize = decltype(std::declval<B_>().globalSize());

    /// Return the number of entries in the local part of the vector
    std::size_t localSize() const
    {
      return Dune::Hybrid::ifElse(Dune::Std::is_detected<HasLocalSize,Backend>{},
        [&](auto id) { return id(backend_).localSize(); },
        [&](auto id) { return id(backend_).size(); });
    }

    /// Return the number of entries in the global vector
    std::size_t globalSize() const
    {
      return Dune::Hybrid::ifElse(Dune::Std::is_detected<HasGlobalSize,Backend>{},
        [&](auto id) { return id(backend_).globalSize(); },
        [&](auto id) { return id(backend_).size(); });
    }

    /// Resize the \ref vector to the size of the \ref basis
    void resize()
    {
      backend_.init(sizeInfo(*basis_), false);
      state_ = VectorState::unknown;
    }

    /// Resize the \ref vector to the size of the \ref basis and set to zero
    void resizeZero()
    {
      backend_.init(sizeInfo(*basis_), true);
      state_ = VectorState::synchronized;
    }

    /// Prepare the vector for insertion of values, finish the insertion with
    /// \ref finish().
    void init(bool clear)
    {
      backend_.init(sizeInfo(*basis_), clear);
      state_ = clear ? VectorState::synchronized : VectorState::unknown;
    }

    /// Finish the insertion of values, see \ref init()
    void finish()
    {
      backend_.finish();
      state_ = VectorState::unknown;
    }

    /// Access the entry \p i of the \ref vector with read-only-access.
    template <class Index>
    typename Backend::value_type DUNE_DEPRECATED_MSG("Do not use element-wise vector access")
    operator[](Index const& idx) const
    {
      assert(state_ == VectorState::synchronized || state_ == VectorState::unknown);
      return at(idx);
    }

    /// Access the entry \p i of the \ref vector with write-access. Uses a proxy
    /// class the redirects to \ref insertValue.
    template <class Index>
    AccessProxy<Index> DUNE_DEPRECATED_MSG("Do not use element-wise vector access")
    operator[](Index const& idx)
    {
      return AccessProxy<Index>{this, idx};
    }

    /// Return the value of the vector at the local index \ref idx
    template <class Index>
    typename Backend::value_type at(Index const& idx) const
    {
      const_cast<Self*>(this)->synchronize();

      return backend_.at(flatMultiIndex(idx));
    }

    /// \brief Insert a single value into the matrix (add or overwrite to existing value)
    /**
     * Inserts or adds a value into a certain location \p dof (given as dof multi-index)
     * of a vector. The insertion mode is determined by the \p assign functor. Use
     * \ref Assigner::plus_assign for adding values (default) or \ref Assigner::assign
     * for overwriting (setting) values. Different insertion modes can not be mixed!
     *
     * Insertion must be closed with a call to \ref finish().
     *
     * [[not collective]]
     */
    template <class Index, class Assign = Assigner::plus_assign>
    void insert(Index const& dof, typename Backend::value_type const& value, Assign assign = {})
    {
      test_exit_dbg(state_ == VectorStateOf_t<Assign>::value ||
                    state_ == VectorState::unknown ||
                    state_ == VectorState::synchronized,
        "Vector in invalid state {} for insertion by {}.", to_string(state_), Dune::className<Assign>());

      backend_.insert(flatMultiIndex(dof), value, assign);

      // set the state to insert_values or add_values
      state_ = VectorStateOf_t<Assign>::value;
    }

    /// See \ref insert for assignment operation \ref Assigner::assign
    template <class Index>
    void set(Index const& dof, typename Backend::value_type const& value)
    {
      insert(dof, value, Assigner::assign{});
    }

    /// See \ref insert for assignment operation \ref Assigner::plus_assign
    template <class Index>
    void add(Index const& dof, typename Backend::value_type const& value)
    {
      insert(dof, value, Assigner::plus_assign{});
    }


    /// \brief Extract values from the vector referring to the given local indices
    /// and store it into a buffer
    /**
     * Collect value of indices and store them into a buffer. The buffer must be
     * a vector-like container with `buffer.resize()` and `buffer.begin()`. The
     * indices must be stored in an iterable container.
     *
     * If the vector is not in synchronized state, collects all necessary values
     * possibly from neighbouring processors.
     *
     * [[expects: localView is bound to an element]]
     * [[expects: node is in localView.tree()]]
     * [[possibly neighbor-wise collective operation]]
     */
    template <class Node, class Buffer,
      REQUIRES(Dune::models<Dune::Functions::Concept::BasisNode, Node>())>
    void gather(LocalView const& localView, Node const& node, Buffer& buffer) const
    {
      test_exit(state_ == VectorState::unknown ||
                state_ == VectorState::synchronized,
        "Vector in invalid state {} for gather operations.", to_string(state_));

      const_cast<Self*>(this)->synchronize();

      buffer.resize(node.size());
      backend_.gather(nodeIndices(localView, node), buffer.begin());
    }

    // [[expects: localView is bound to an element]]
    template <class Buffer>
    void gather(LocalView const& localView, Buffer& buffer) const
    {
      gather(localView, localView.tree(), buffer);
    }

    /// Insert a block of values into the vector (add or overwrite to existing values)
    /**
     * Inserts or adds values into certain locations of a vector. Insertion indices
     * are extracted from the given \p localView. The insertion mode is determined
     * by the \p assign functor. Use \ref Assigner::plus_assign for adding values
     * (default) or \ref Assigner::assign for overwriting (setting) values. Different
     * insertion modes can not be mixed! The \p localVector is assumed to be a continuous
     * memory container with a `data()` method to get a pointer to the beginning.
     *
     * The \p mask models a boolean range with at least a `begin()` method. Must
     * be forward iterable for at least `localVector.size()` elements. Does not
     * need an `end()` method. See, e.g. \ref FakeContainer.
     *
     * Insertion must be closed with a call to \ref finish(). It is not allowed
     * to switch insertion mode before calling `finish()`.
     *
     * [[expects: localView is bound to an element]]
     * [[expects: node is in localView.tree()]]
     * [[not collective]]
     */
    template <class Node, class NodeVector, class MaskRange, class Assign,
      REQUIRES(Dune::models<Dune::Functions::Concept::BasisNode, Node>())>
    void scatter(LocalView const& localView, Node const& node, NodeVector const& localVector,
                 MaskRange const& mask, Assign assign)
    {
      test_exit(state_ == VectorStateOf_t<Assign>::value ||
                state_ == VectorState::unknown ||
                state_ == VectorState::synchronized,
        "Vector in invalid state {} for insertion by {}.", to_string(state_), Dune::className<Assign>());

      assert(localVector.size() == node.size());

      // create a range of DOF indices on the node
      backend_.scatter(nodeIndices(localView, node), localVector, mask, assign);

      // set the state to insert_values or add_values
      state_ = VectorStateOf_t<Assign>::value;
    }

    /// Call \ref scatter with `MaskRange` set to \ref FakeContainer.
    // [[expects: localView is bound to an element]]
    // [[expects: node is in localView.tree()]]
    template <class Node, class NodeVector, class Assign,
      REQUIRES(Dune::models<Dune::Functions::Concept::BasisNode, Node>())>
    void scatter(LocalView const& localView, Node const& node, NodeVector const& localVector, Assign assign)
    {
      scatter(localView, node, localVector, FakeContainer<bool,true>{}, assign);
    }

    /// Call \ref scatter with `Node` given by the tree of the \p localView.
    // [[expects: localView is bound to an element]]
    template <class LocalVector, class Assign>
    void scatter(LocalView const& localView, LocalVector const& localVector, Assign assign)
    {
      scatter(localView, localView.tree(), localVector, assign);
    }

    /// Copies a block of values from \p vector to this
    /**
     * The values to copy might be masked with the range \p mask.
     * The assignment mode \p assign must be one of \ref Assigner::assign, or
     * \ref Assigner::plus_Assign.
     *
     * Requires the (local) size of the vector to be equal to the basis dimension.
     *
     * [[not collective]]
     **/
    template <class Vector, class MaskRange, class Assign>
    void copy(Vector const& vector, MaskRange const& mask, Assign assign)
    {
      test_exit(state_ == VectorStateOf_t<Assign>::value ||
                state_ == VectorState::unknown ||
                state_ == VectorState::synchronized,
        "Vector in invalid state {} for insertion by {}.", to_string(state_), Dune::className<Assign>());

      using size_type = typename Backend::size_type;
      size_type size = sizeInfo(*basis_);
      backend_.scatter(Dune::range(size), vector, mask, assign);

      // set the state to insert_values or add_values
      state_ = VectorStateOf_t<Assign>::value;
    }

    /// Call \ref copy with `MaskRange` set to \ref FakeContainer.
    template <class Vector, class Assign>
    void copy(Vector const& vector, Assign assign)
    {
      copy(vector, FakeContainer<bool,true>{}, assign);
    }

    /// Apply \p func to each value at given indices \p localInd
    /**
     * First, synchronizes the values of the vector, then applies the functor to
     * each local value associated to the local indices \p localInd.
     *
     * [[mutable]]
     **/
    template <class LocalInd, class Func>
    void forEach(LocalInd const& localInd, Func&& func)
    {
      synchronize();
      backend_.forEach(localInd, FWD(func));
    }

    /// Call \ref forEach for all entries in the vector.
    template <class Func>
    void forEach(Func&& func)
    {
      using size_type = typename Backend::size_type;
      size_type size = sizeInfo(*basis_);
      forEach(Dune::range(size), FWD(func));
    }

    /// Apply \p func to each value at given indices \p localInd
    /**
     * First, synchronizes the values of the vector, then applies the functor to
     * each local value associated to the local indices \p localInd.
     *
     * [[const]]
     **/
    template <class LocalInd, class Func>
    void forEach(LocalInd const& localInd, Func&& func) const
    {
      const_cast<Self*>(this)->synchronize();
      backend_.forEach(localInd, FWD(func));
    }

    /// Call \ref forEach for all entries in the vector.
    template <class Func>
    void forEach(Func&& func) const
    {
      using size_type = typename Backend::size_type;
      size_type size = sizeInfo(*basis_);
      forEach(Dune::range(size), FWD(func));
    }


  private:
    // synchronizes ghost values. Does not touch owned values
    void synchronize()
    {
      if (state_ != VectorState::synchronized)
        backend_.synchronize();

      state_ = VectorState::synchronized;
    }

    // print the vector state to string for debugging purposes
    static std::string to_string(VectorState state)
    {
      switch (state) {
        case VectorState::synchronized: return "synchronized";
        case VectorState::insert_values: return "insert_values";
        case VectorState::add_values: return "add_values";
        default: return "unknown";
      }
    }

  private:
    /// The finite element space / basis associated with the data vector
    std::shared_ptr<GlobalBasis> basis_;

    /// Data backend
    Backend backend_;

    /// The current state of the vector, one of {synchronized, insert_values,
    /// add_values, unknown}
    VectorState state_ = VectorState::unknown;
  };


  template <class GB, class B>
    template <class Index>
  struct VectorBase<GB,B>::AccessProxy
  {
    using value_type = typename Backend::value_type;

    void operator=(value_type const& value)
    {
      self->set(i, value);
    }

    void operator+=(value_type const& value)
    {
      self->add(i, value);
    }

    void operator-=(value_type const& value)
    {
      self->add(i, -value);
    }

    operator value_type() const
    {
      return self->at(i);
    }

    VectorBase* self;
    Index i;
  };

} // end namespace AMDiS
