#pragma once

#include <Eigen/Dense>

#include <dune/common/ftraits.hh>

#include <amdis/Output.hpp>
#include <amdis/common/FakeContainer.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector and a corresponding basis
  template <class T>
  class VectorBackend
  {
  public:
    using Traits = T;

    /// The type of the base vector
    using BaseVector = typename Traits::Vec;

    /// The type of the elements of the DOFVector
    using value_type = typename BaseVector::Scalar;

    /// The type of the elements of the DOFVector
    using block_type = value_type;

    /// The underlying field type
    using field_type = typename Dune::FieldTraits<value_type>::field_type;

    /// The index/size - type
    using size_type  = typename BaseVector::Index;

  public:
    /// Constructor. Constructs new BaseVector.
    template <class Basis>
    explicit VectorBackend(Basis const&) {}

    /// Return the data-vector \ref vector_
    BaseVector const& vector() const
    {
      return vector_;
    }

    /// Return the data-vector \ref vector_
    BaseVector& vector()
    {
      return vector_;
    }

    /// Return the current size of the \ref vector_
    std::size_t size() const
    {
      return vector_.size();
    }


    /// Resize the \ref vector_ to the size \p s
    template <class SizeInfo>
    void init(SizeInfo const& size, bool clear)
    {
      vector_.resize(size_type(size));
      if (clear)
        vector_.setZero();
    }

    void finish() { /* do nothing */ }
    void synchronize() { /* do nothing */ }


    /// Access the entry \p i of the \ref vector with read-access.
    value_type const& at(size_type i) const
    {
      test_exit_dbg(i < vector_.size(),
        "Index {} out of range [0,{})", i, vector_.size());
      return vector_.coeff(i);
    }

    /// Access the entry \p i of the \ref vector with write-access.
    template <class Assign>
    void insert(size_type i, value_type value, Assign assign)
    {
      test_exit_dbg(i < vector_.size(),
        "Index {} out of range [0,{})", i, vector_.size());
      assign(value, vector_.coeffRef(i));
    }

    template <class IndexRange, class OutputIterator>
    void gather(IndexRange const& localInd, OutputIterator buffer) const
    {
      for (size_type i : localInd)
        *(buffer++) = vector_.coeff(i);
    }

    template <class IndexRange, class LocalVec, class Assign>
    void scatter(IndexRange const& localInd, LocalVec const& vec, FakeContainer<bool,true>, Assign assign)
    {
      auto vec_it = std::begin(vec);
      for (size_type i : localInd)
        assign(*vec_it++, vector_.coeffRef(i));
    }

    template <class IndexRange, class LocalVec, class MaskRange, class Assign>
    void scatter(IndexRange const& localInd, LocalVec const& vec, MaskRange const& mask, Assign assign)
    {
      auto vec_it = std::begin(vec);
      auto mask_it = std::begin(mask);
      auto ind_it = std::begin(localInd);
      for (; vec_it != std::end(vec); ++vec_it, ++mask_it, ++ind_it) {
        if (*mask_it)
          assign(*vec_it, vector_.coeffRef(*ind_it));
      }
    }

    template <class IndexRange, class Func>
    void forEach(IndexRange const& localInd, Func&& f) const
    {
      for (size_type i : localInd)
        f(i, vector_.coeff(i));
    }

    template <class IndexRange, class Func>
    void forEach(IndexRange const& localInd, Func&& f)
    {
      for (size_type i : localInd)
        f(i, vector_.coeffRef(i));
    }

  private:
    /// The data-vector (can hold a new BaseVector or a pointer to external data
    BaseVector vector_;
  };

} // end namespace AMDiS
