#pragma once

#include <memory>
#include <utility>

#include <dune/common/typeutilities.hh>
#include <dune/istl/operators.hh>
#include <dune/istl/scalarproducts.hh>
#include <dune/istl/solvercategory.hh>

#if HAVE_MPI
#include <dune/istl/novlpschwarz.hh>
#include <dune/istl/schwarz.hh>
#endif

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/istl/PreconWrapper.hpp>

namespace AMDiS
{
  /// Creator to create ScalarProduct objects
  template <class Vec>
  class ISTLScalarProductCreator
  {
  public:
    using Interface = Dune::ScalarProduct<Vec>;

    template <class Comm>
    static std::unique_ptr<Interface> create(Dune::SolverCategory::Category cat, Comm const& comm)
    {
      return create_impl(cat, comm.get(), Dune::PriorityTag<10>{});
    }

  private:
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, Dune::Amg::SequentialInformation, Dune::PriorityTag<2>)
    {
      assert(cat == Dune::SolverCategory::sequential);
      return std::make_unique< Dune::SeqScalarProduct<Vec> >();
    }

    template <class Communication>
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, Communication const& comm, Dune::PriorityTag<1>)
    {
      switch (cat)
      {
        case Dune::SolverCategory::sequential:
          return std::make_unique< Dune::SeqScalarProduct<Vec> >();
#if HAVE_MPI
        case Dune::SolverCategory::overlapping:
          return std::make_unique< Dune::OverlappingSchwarzScalarProduct<Vec,Communication> >(comm);
        case Dune::SolverCategory::nonoverlapping:
          return std::make_unique< Dune::NonoverlappingSchwarzScalarProduct<Vec,Communication> >(comm);
#endif
        default:
          error_exit("Invalid solver category for scalar product\n");
          return nullptr; // avoid warnings
      }
    }
  };

  /// Creator to create Linear Operator objects
  template <class Mat, class Vec>
  class ISTLLinearOperatorCreator
  {
  public:
    using Interface = Dune::AssembledLinearOperator<Mat,Vec,Vec>;

    template <class Comm>
    static std::unique_ptr<Interface> create(Dune::SolverCategory::Category cat, Mat const& mat, Comm const& comm)
    {
      return create_impl(cat, mat, comm.get(), Dune::PriorityTag<10>{});
    }

  private:
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, Mat const& mat, Dune::Amg::SequentialInformation, Dune::PriorityTag<2>)
    {
      assert(cat == Dune::SolverCategory::sequential);
      return std::make_unique< Dune::MatrixAdapter<Mat,Vec,Vec> >(mat);
    }

    template <class Communication>
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, Mat const& mat, Communication const& comm, Dune::PriorityTag<1>)
    {
      switch (cat)
      {
        case Dune::SolverCategory::sequential:
          return std::make_unique< Dune::MatrixAdapter<Mat,Vec,Vec> >(mat);
#if HAVE_MPI
        case Dune::SolverCategory::overlapping:
          return std::make_unique< Dune::OverlappingSchwarzOperator<Mat,Vec,Vec,Communication> >(mat, comm);
        case Dune::SolverCategory::nonoverlapping:
          return std::make_unique< Dune::NonoverlappingSchwarzOperator<Mat,Vec,Vec,Communication> >(mat, comm);
#endif
        default:
          error_exit("Invalid solver category for linear operator\n");
          return nullptr; // avoid warnings
      }
    }
  };


  /// Creator to create parallel Preconditioners
  template <class Vec>
  class ISTLParallelPreconditionerCreator
  {
  public:
    using Interface = Dune::Preconditioner<Vec,Vec>;

    template <class Comm>
    static std::unique_ptr<Interface> create(Dune::SolverCategory::Category cat, std::unique_ptr<Interface> prec, Comm const& comm)
    {
      return create_impl(cat, std::move(prec), comm.get(), Dune::PriorityTag<10>{});
    }

  private:
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, std::unique_ptr<Interface> prec, Dune::Amg::SequentialInformation, Dune::PriorityTag<2>)
    {
      assert(cat == Dune::SolverCategory::sequential);
      return std::move(prec);
    }

    template <class Communication>
    static std::unique_ptr<Interface>
    create_impl(Dune::SolverCategory::Category cat, std::unique_ptr<Interface> prec, Communication const& comm, Dune::PriorityTag<1>)
    {
      switch (cat)
      {
        case Dune::SolverCategory::sequential:
          return std::move(prec);
#if HAVE_MPI
        case Dune::SolverCategory::overlapping:
        {
          using BP = Dune::BlockPreconditioner<Vec,Vec,Communication>;
          return std::make_unique< PreconWrapper<BP,Interface> >(std::move(prec), comm);
        }
        case Dune::SolverCategory::nonoverlapping:
        {
          using BP = Dune::NonoverlappingBlockPreconditioner<Communication,Interface>;
          return std::make_unique< PreconWrapper<BP,Interface> >(std::move(prec), comm);
        }
#endif
        default:
          error_exit("Invalid solver category for parallel preconditioner\n");
          return nullptr; // avoid warnings
      }
    }
  };

} // end namespace AMDiS
