#pragma once

// MTL4 headers
#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/itl/pc/ilu_0.hpp>
#include <boost/numeric/itl/pc/ic_0.hpp>
#include <boost/numeric/mtl/vector/assigner.hpp>

#include <amdis/linearalgebra/mtl/itl/masslumping.hpp>
#include <amdis/linearalgebra/mtl/Preconditioner.hpp>
#include <amdis/linearalgebra/mtl/Traits.hpp>
#include <amdis/CreatorMap.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   * \class AMDiS::DiagonalPreconditioner
   * \brief ITL_Preconditioner implementation of diagonal (jacobi) preconditioner,
   * \implements ITL_Preconditioner
   *
   * Diagonal preconditioner \f$ M^{-1} \f$ for the system \f$ Ax=b \f$ is defined as: \f$ M=diag(A) \f$.
   */
  template <class Matrix>
  using DiagonalPreconditioner = itl::pc::diagonal<Matrix>;

  /**
   * \ingroup Solver
   * \class AMDiS::DiagonalPreconditioner
   * \brief ITL_Preconditioner implementation of diagonal (jacobi) preconditioner,
   * \implements ITL_Preconditioner
   *
   * Diagonal preconditioner \f$ M^{-1} \f$ for the system \f$ Ax=b \f$ is defined as: \f$ M_ii=sum_j(A_ij) \f$.
   */
  template <class Matrix>
  using MassLumpingPreconditioner = itl::pc::masslumping<Matrix>;


  /**
   * \ingroup Solver
   * \class AMDiS::IdentityPreconditioner
   * \brief ITL_Preconditioner implementation of identity preconditioner,
   * \implements ITL_Preconditioner
   *
   * Identity preconditioner. Behaves like no preconditioning.
   */
  template <class Matrix>
  using IdentityPreconditioner = itl::pc::identity<Matrix>;


  /**
   * \ingroup Solver
   * \class AMDiS::ILUPreconditioner
   * \brief ITL_Preconditioner implementation of ILU (Incomplete LU factorization)
   * preconditioner. \implements ITL_Preconditioner
   *
   * The preconditioner is used from ITL. It corresponds for instance to
   * "Iterative Methods for Sparce Linear Systems", second edition, Yousef Saad.
   *  The preconditioner is described in chapter 10.3 (algorithm 10.4).
   */
  template <class Matrix>
  using ILUPreconditioner = itl::pc::ilu_0<Matrix>;


  /**
   * \ingroup Solver
   * \class AMDiS::ICPreconditioner
   * \brief ITL_Preconditioner implementation of IC (Incomplete Cholesky factorization)
   * preconditioner. \implements ITL_Preconditioner
   *
   * IC (Incomplete Cholesky factorization) preconditioner.
   */
  template <class Matrix>
  using ICPreconditioner = itl::pc::ic_0<Matrix>;

  template <class Traits>
  class DefaultCreators< PreconditionerInterface<Traits> >
  {
    using Matrix = typename Traits::Mat;
    using PreconBase = PreconditionerInterface<Traits>;

    template <template<class> class ITLPrecon>
    using PreconCreator
      = typename Preconditioner<Traits, ITLPrecon<Matrix>>::Creator;

    using Map = CreatorMap<PreconBase>;

  public:
    static void init()
    {
      auto pc_diag = new PreconCreator<DiagonalPreconditioner>;
      Map::addCreator("diag", pc_diag);
      Map::addCreator("jacobi", pc_diag);

      auto pc_mass = new PreconCreator<MassLumpingPreconditioner>;
      Map::addCreator("masslumping", pc_mass);

      auto pc_ilu = new PreconCreator<ILUPreconditioner>;
      Map::addCreator("ilu", pc_ilu);
      Map::addCreator("ilu0", pc_ilu);

      auto pc_ic = new PreconCreator<ICPreconditioner>;
      Map::addCreator("ic", pc_ic);
      Map::addCreator("ic0", pc_ic);

      auto pc_id = new PreconCreator<IdentityPreconditioner>;
      Map::addCreator("identity", pc_id);
      Map::addCreator("no", pc_id);

      Map::addCreator("default", pc_id);
    }
  };

  template <class Traits>
  itl::pc::solver<PreconditionerInterface<Traits>, typename Traits::Vec, false>
  solve(PreconditionerInterface<Traits> const& P, typename Traits::Vec const& vin)
  {
    return {P, vin};
  }

  template <class Traits>
  itl::pc::solver<PreconditionerInterface<Traits>, typename Traits::Vec, true>
  adjoint_solve(PreconditionerInterface<Traits> const& P, typename Traits::Vec const& vin)
  {
    return {P, vin};
  }

} // namespace AMDiS
