#pragma once

#ifdef HAVE_UMFPACK

#include <algorithm>
#include <string>

#include <boost/numeric/mtl/operation/two_norm.hpp>
#include <boost/numeric/mtl/interface/umfpack_solve.hpp>

#include <amdis/linearalgebra/mtl/Traits.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/Output.hpp>

namespace AMDiS
{
  /**
  * \ingroup Solver
  * \class AMDiS::UmfpackRunner
  * \brief \implements RunnerInterface
  * Wrapper for the external
  *   [UMFPACK solver](http://faculty.cse.tamu.edu/davis/suitesparse.html)
  *
  * This is a direct solver for large sparse matrices.
  */
  template <class Traits>
  class UmfpackRunner
      : public RunnerInterface<Traits>
  {
    using Matrix = typename Traits::Mat;
    using Vector = typename Traits::Vec;
    using Comm = typename Traits::Comm;
    using Super = RunnerInterface<Traits>;

    using SolverType = mtl::mat::umfpack::solver<Matrix>;

  public:
    /// Constructor. Reads UMFPACK parameters from initfile
    UmfpackRunner(std::string const& prefix)
    {
      Parameters::get(prefix + "->store symbolic", storeSymbolic_);
      Parameters::get(prefix + "->symmetric strategy", symmetricStrategy_);
      Parameters::get(prefix + "->alloc init", allocInit_);
    }

    /// Implementation of \ref RunnerInterface::init()
    void init(Matrix const& matrix, Comm&) override
    {
      try {
        if (bool(solver_) && storeSymbolic_)
          solver_->update_numeric();
        else
          solver_.reset(new SolverType(matrix, symmetricStrategy_, allocInit_));
      } catch (mtl::mat::umfpack::error const& e) {
        umfpack_error_exit("factorize", e.code);
      }
    }

    /// Implementation of \ref RunnerInterface::exit()
    void exit() override {}

    /// Implementation of \ref RunnerBase::solve()
    int solve(Matrix const& A, Vector& x, Vector const& b, SolverInfo& solverInfo) override
    {
      test_exit(bool(solver_), "The umfpack solver was not initialized\n");

      int code = 0;
      try {
        code = (*solver_)(x, b);
      } catch (mtl::mat::umfpack::error& e) {
        umfpack_error_exit("solve", e.code);
      }

      auto r = Vector(b);
      if (two_norm(x) != 0)
        r -= A * x;

      solverInfo.setAbsResidual(two_norm(r));
      solverInfo.setError(code);

      return code;
    }

    static void umfpack_error_exit(std::string solutionPhase, int code)
    {
      static std::map<int, std::string> error_message = {
        {UMFPACK_OK, "UMFPACK was successful"},
        {UMFPACK_WARNING_singular_matrix, "Matrix is singular. There are exact zeros on the diagonal."},
        {UMFPACK_WARNING_determinant_underflow, "The determinant is nonzero, but smaller in magnitude than the smallest positive floating-point number."},
        {UMFPACK_WARNING_determinant_overflow, "The determinant is larger in magnitude than the largest positive floating-point number (IEEE Inf)."},
        {UMFPACK_ERROR_out_of_memory, "Not enough memory. The ANSI C malloc or realloc routine failed."},
        {UMFPACK_ERROR_invalid_Numeric_object, "Invalid Numeric object."},
        {UMFPACK_ERROR_invalid_Symbolic_object, "Invalid Symbolic object."},
        {UMFPACK_ERROR_argument_missing, "Some required arguments are missing."},
        {UMFPACK_ERROR_n_nonpositive, "The number of rows or columns of the matrix must be greater than zero."},
        {UMFPACK_ERROR_invalid_matrix, "The matrix is invalid."},
        {UMFPACK_ERROR_different_pattern, "The pattern of the matrix has changed between the symbolic and numeric factorization."},
        {UMFPACK_ERROR_invalid_system, "The sys argument provided to one of the solve routines is invalid."},
        {UMFPACK_ERROR_invalid_permutation, "The permutation vector provided as input is invalid."},
        {UMFPACK_ERROR_file_IO, "Error in file IO"},
        {UMFPACK_ERROR_ordering_failed, "The ordering method failed."},
        {UMFPACK_ERROR_internal_error, "An internal error has occurred, of unknown cause."}
      };

      error_exit("UMFPACK_ERROR({}, {}) = {}", solutionPhase, code, error_message[code]);
    }

  protected:
    std::shared_ptr<SolverType> solver_;

    bool storeSymbolic_ = false;
    int symmetricStrategy_ = 0;
    double allocInit_ = 0.7;
  };

} // end namespace AMDiS

#endif // HAVE_UMFPACK
