#pragma once

#include <boost/numeric/itl/pc/diagonal.hpp>

#include <amdis/linearalgebra/mtl/BlockMTLMatrix.hpp>

namespace itl
{
  namespace pc
  {
    /// Diagonal Preconditioner
    template <class MTLMatrix, std::size_t _N, std::size_t _M, class Value>
    class diagonal<AMDiS::BlockMTLMatrix<MTLMatrix, _N, _M>, Value>
    {
      using Self       = diagonal;
      using MatrixType = AMDiS::BlockMTLMatrix<MTLMatrix, _N, _M>;

    public:
      using value_type = Value;
      using size_type  = typename mtl::Collection<MatrixType>::size_type;

      /// Constructor takes matrix reference
      explicit diagonal(const MatrixType& A) : inv_diag(num_rows(A))
      {
        MTL_THROW_IF(num_rows(A) != num_cols(A), mtl::matrix_not_square());
        using math::reciprocal;

        std::array<mtl::irange, _N> r_rows;
        A.getRowRanges(r_rows);

        for (std::size_t i = 0; i < _N; i++)
          inv_diag[r_rows[i]] = mtl::mat::diagonal(A[i][i]);

        for (size_type i = 0; i < num_rows(A); ++i)
          inv_diag[i]= reciprocal(inv_diag[i]);
      }

      /// Member function solve, better use free function solve
      template <typename Vector>
      Vector solve(const Vector& x) const
      {
        Vector y(resource(x));
        solve(x, y);
        return y;
      }

      template <typename VectorIn, typename VectorOut>
      void solve(const VectorIn& x, VectorOut& y) const
      {
        y.checked_change_resource(x);
        MTL_THROW_IF(size(x) != size(inv_diag), mtl::incompatible_size());
        for (size_type i= 0; i < size(inv_diag); ++i)
          y[i]= inv_diag[i] * x[i];
      }

      /// Member function for solving adjoint problem, better use free function adjoint_solve
      template <typename Vector>
      Vector adjoint_solve(const Vector& x) const
      {
        Vector y(resource(x));
        adjoint_solve(x, y);
        return y;
      }

      template <typename VectorIn, typename VectorOut>
      void adjoint_solve(const VectorIn& x, VectorOut& y) const
      {
        using mtl::conj;
        y.checked_change_resource(x);
        MTL_THROW_IF(size(x) != size(inv_diag), mtl::incompatible_size());
        for (size_type i= 0; i < size(inv_diag); ++i)
          y[i]= conj(inv_diag[i]) * x[i];
      }

    protected:
      mtl::vec::dense_vector<value_type>    inv_diag;
    };


  }
} // namespace itl::pc
