#pragma once

#include <petscvec.h>

#include <dune/common/ftraits.hh>
#include <dune/common/shared_ptr.hh>

#include <amdis/Output.hpp>
#include <amdis/common/FakeContainer.hpp>
#include <amdis/functions/Interpolate.hpp>
#include <amdis/operations/Assigner.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector data
  template <class T>
  class VectorBackend
  {
    using Communication = typename T::Comm;

    template <class A>
    using AssignMode = std::conditional_t<std::is_same<A,Assigner::plus_assign>::value,
      std::integral_constant<::InsertMode, ADD_VALUES>,
      std::integral_constant<::InsertMode, INSERT_VALUES>>;

  public:
    using Traits = T;

    /// The type of the base vector
    using BaseVector = ::Vec;

    /// The type of the elements of the DOFVector
    using value_type = PetscScalar;

    /// The index/size - type
    using size_type  = PetscInt;

  public:
    /// Constructor. Constructs new BaseVector.
    template <class Basis>
    explicit VectorBackend(Basis const& basis)
      : comm_(Dune::stackobject_to_shared_ptr(basis.communicator()))
    {}

    /// Move constructor
    VectorBackend(VectorBackend&& other)
    {
      swap(*this, other);
    }

    /// Copy constructor
    VectorBackend(VectorBackend const& other)
      : comm_(other.comm_)
      , initialized_(other.initialized_)
    {
      if (other.initialized_) {
        VecDuplicate(other.vector_, &vector_);
        VecCopy(other.vector_, vector_);
      }
    }

    /// Destructor
    ~VectorBackend()
    {
      destroy();
    }

    /// Move assignment operator
    VectorBackend& operator=(VectorBackend&& other)
    {
      swap(*this, other);
      return *this;
    }

    /// Copy assignment operator
    VectorBackend& operator=(VectorBackend const& other)
    {
      return *this = VectorBackend(other);
    }

    /// Return the data-vector \ref vector_
    BaseVector const& vector() const
    {
      return vector_;
    }

    /// Return the data-vector \ref vector_
    BaseVector& vector()
    {
      return vector_;
    }

    /// Return the number of entries in the local part of the vector
    std::size_t localSize() const
    {
      return dofMap().localSize();
    }

    /// Return the number of entries in the global vector
    std::size_t globalSize() const
    {
      return dofMap().globalSize();
    }

    /// Resize the \ref vector_ to the size \p s
    // [[possibly collective]]
    template <class SizeInfo>
    void init(SizeInfo const&, bool clear)
    {
      Dune::Timer t;
      if (!initialized_) {
        // create the vector
        create();
        info(3, "  create new vector needed {} seconds", t.elapsed());
        t.reset();
      }
      else {
        PetscInt localSize = 0, globalSize = 0;
        VecGetSize(vector_, &globalSize);
        VecGetLocalSize(vector_, &localSize);

        // re-create the vector
        if (std::size_t(localSize) != dofMap().localSize() ||
            std::size_t(globalSize) != dofMap().globalSize()) {
          destroy();
          create();
          info(3, "  destroy old and create new vector needed {} seconds", t.elapsed());
          t.reset();
        }
      }

      if (clear)
        setZero();

      initialized_ = true;
    }

    /// Finish assembly. Must be called before vector can be used in a KSP
    // [[collective]]
    void finish()
    {
      Dune::Timer t;
      VecAssemblyBegin(vector_);
      VecAssemblyEnd(vector_);
      info(3, "  finish vector assembly needed {} seconds", t.elapsed());
    }

    /// Update the ghost regions with values from the owning process
    // [[collective]]
    void synchronize()
    {
      Dune::Timer t;
      VecGhostUpdateBegin(vector_, INSERT_VALUES, SCATTER_FORWARD);
      VecGhostUpdateEnd(vector_, INSERT_VALUES, SCATTER_FORWARD);
      info(3, "  synchronize vector needed {} seconds", t.elapsed());
    }

    /// Access the entry \p i of the \ref vector with read-only-access.
    // [[not collective]]
    PetscScalar at(std::size_t i) const
    {
      PetscScalar y = 0;
      gather(std::array<std::size_t,1>{i}, &y);
      return y;
    }

    /// Access the entry \p i of the \ref vector with write-access.
    // [[not collective]]
    template <class Assign>
    void insert(std::size_t dof, PetscScalar value, Assign)
    {
      assert(initialized_);
      VecSetValue(vector_, dofMap().global(dof), value, AssignMode<Assign>::value);
    }

    /// Collect all values from this vector to indices given by `localInd` and store it into the output buffer.
    // [[not collective]]
    template <class IndexRange, class OutputIterator>
    void gather(IndexRange const& localInd, OutputIterator buffer) const
    {
      forEach(localInd, [&buffer](auto&&, auto const& value) { *buffer = value; ++buffer; });
    }

    /// Store all values given by `localVal` with indices `localInd` in the vector
    /**
     * Store only those values with mask != 0. Use an assign mode for the storage. It is not allowed
     * to switch the assign mode before calling VecAssemblyBegin and VecAssemblyEnd.
     **/
    // [[not collective]]
    template <class IndexRange, class LocalValues, class MaskRange, class Assign>
    void scatter(IndexRange const& localInd, LocalValues const& localVal, MaskRange const& mask, Assign)
    {
      static_assert(std::is_same<typename LocalValues::value_type, PetscScalar>::value, "");
      assert(initialized_);
      assert(localInd.size() == localVal.size());

      // map local to global indices
      std::vector<PetscInt> globalInd;
      globalInd.reserve(localInd.size());

      auto ind_it = std::begin(localInd);
      auto mask_it = std::begin(mask);
      for (; ind_it != std::end(localInd); ++mask_it, ++ind_it)
        globalInd.push_back(*mask_it ? dofMap().global(*ind_it) : -1);

      if (! std::is_same<MaskRange, FakeContainer<bool,true>>::value)
        VecSetOption(vector_, VEC_IGNORE_NEGATIVE_INDICES, PETSC_TRUE);

      VecSetValues(vector_, globalInd.size(), globalInd.data(), localVal.data(), AssignMode<Assign>::value);
      VecSetOption(vector_, VEC_IGNORE_NEGATIVE_INDICES, PETSC_FALSE);
    }

    /// Apply the functor `f` to each vector entry to local index in `localInd` range.
    // [[not collective]]
    template <class IndexRange, class Func>
    void forEach(IndexRange const& localInd, Func&& f) const
    {
      // Obtains the local ghosted representation of a parallel vector
      Vec localGhostedVec = nullptr;
      VecGhostGetLocalForm(vector_, &localGhostedVec);
      assert(localGhostedVec != nullptr);

      // A pointer to a contiguous array that contains this processor's portion of the vector data.
      PetscScalar *ptr;
      VecGetArray(localGhostedVec, &ptr);

      for (auto const& dof : localInd)
      {
        if (dofMap().owner(dof)) {
          // the index is a purely local entry
          f(dof, ptr[dofMap().dofToLocal(dof)]);
        }
        else {
          // ghost entries are stored at the end of the array
          f(dof, ptr[dofMap().localSize() + dofMap().dofToGhost(dof)]);
        }
      }

      VecRestoreArray(localGhostedVec, &ptr);
      VecGhostRestoreLocalForm(vector_, &localGhostedVec);
    }


    /// Set all entries to \p value, including the ghost entries
    // [[not collective]]
    void set(PetscScalar value)
    {
      Vec localGhostedVec = nullptr;
      VecGhostGetLocalForm(vector_, &localGhostedVec);
      assert(localGhostedVec != nullptr);

      VecSet(localGhostedVec, value);
      VecGhostRestoreLocalForm(vector_, &localGhostedVec);
    }

    /// Zero all entries, including the ghost entries
    // [[not collective]]
    void setZero()
    {
      Vec localGhostedVec = nullptr;
      VecGhostGetLocalForm(vector_, &localGhostedVec);
      assert(localGhostedVec != nullptr);

      VecZeroEntries(localGhostedVec);
      VecGhostRestoreLocalForm(vector_, &localGhostedVec);
    }

    friend void swap(VectorBackend& lhs, VectorBackend& rhs)
    {
      std::swap(lhs.comm_, rhs.comm_);
      std::swap(lhs.vector_, rhs.vector_);
      std::swap(lhs.initialized_, rhs.initialized_);
    }

  private:
    // The local-to-global map
    auto const& dofMap() const
    {
      return comm_->dofMap();
    }

    // An MPI Communicator or PETSC_COMM_SELF
    MPI_Comm comm() const
    {
#if HAVE_MPI
      return Environment::comm();
#else
      return PETSC_COMM_SELF;
#endif
    }

    // Create a new Vector
    void create()
    {
      VecCreateGhost(comm(),
        dofMap().localSize(), dofMap().globalSize(), dofMap().ghostSize(), dofMap().ghostIndices().data(),
        &vector_);
    }

    // Destroy an old vector if created before
    void destroy()
    {
      if (initialized_)
        VecDestroy(&vector_);

      initialized_ = false;
    }

  private:
    std::shared_ptr<Communication const> comm_;

    /// The data-vector
    mutable Vec vector_ = nullptr;
    bool initialized_ = false;
  };

} // end namespace AMDiS
