#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/utility/LocalBasisCache.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct divtestvec {};
  }


  /// first-order operator \f$ \langle\nabla\cdot\Psi, c\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::divtestvec, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::divtestvec, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::divtestvec, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class CG, class Node, class Vec>
    void getElementVector(CG const& context, Node const& node, Vec& elementVector)
    {
      static_assert(Node::isPower, "Node must be a Power-Node.");

      static const std::size_t CHILDREN = Node::CHILDREN;
      static_assert( CHILDREN == CG::dow, "" );

      auto const& quad = this->getQuadratureRule(context.type(), node);
      auto const& localFE = node.child(0).finiteElement();
      std::size_t feSize = localFE.size();

      using RangeFieldType = typename NodeQuadCache<typename Node::ChildType>::Traits::RangeFieldType;
      NodeQuadCache<typename Node::ChildType> cache(localFE.localBasis());

      auto const& shapeGradientsCache = cache.evaluateJacobianAtQP(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,CG::dow>;
        thread_local std::vector<WorldVector> gradients;
        gradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t j = 0; j < feSize; ++j) {
          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_kj = node.child(k).localIndex(j);
            elementVector[local_kj] += factor * gradients[j][k];
          }
        }
      }

    }
  };

  /** @} **/

} // end namespace AMDiS
