#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trial {};
  }


  /// zero-order operator \f$ \langle\psi, c\,\phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::test_trial, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trial, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_trial, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      const bool sameFE = std::is_same<FiniteElementType_t<RN>, FiniteElementType_t<CN>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(contextGeo, quad, rowNode, colNode, elementMatrix);
      else
        getElementMatrixStandard(contextGeo, quad, rowNode, colNode, elementMatrix);
    }

    template <class CG, class Node, class Vec>
    void getElementVector(CG const& contextGeo, Node const& node, Vec& elementVector)
    {
      static_assert(Node::isLeaf, "Operator can be applied to Leaf-Nodes only");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), node);
      auto const& localFE = node.finiteElement();
      std::size_t size = localFE.size();

      NodeQuadCache<Node> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position())
                                                      * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];
        for (std::size_t i = 0; i < size; ++i) {
          const auto local_i = node.localIndex(i);
          elementVector[local_i] += factor * shapeValues[i];
        }
      }
    }


  protected:
    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixStandard(CG const& contextGeo, QR const& quad,
                                  RN const& rowNode, CN const& colNode,
                                  Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<CN> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto local_i = rowNode.localIndex(i);
          const auto value = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colSize; ++j) {
            const auto local_j = colNode.localIndex(j);
            elementMatrix[local_i][local_j] += value * colShapeValues[j];
          }
        }
      }
    }


    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixOptimized(CG const& contextGeo, QR const& quad,
                                   RN const& node, CN const& /*colNode*/,
                                   Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& localFE = node.finiteElement();
      std::size_t size = localFE.size();

      NodeQuadCache<RN> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];

        for (std::size_t i = 0; i < size; ++i) {
          const auto local_i = node.localIndex(i);

          const auto value = factor * shapeValues[i];
          elementMatrix[local_i][local_i] += value * shapeValues[i];

          for (std::size_t j = i+1; j < size; ++j) {
            const auto local_j = node.localIndex(j);

            elementMatrix[local_i][local_j] += value * shapeValues[j];
            elementMatrix[local_j][local_i] += value * shapeValues[j];
          }
        }
      }
    }
  };


  /// Create a zero-order term
  template <class Expr, class... QuadratureArgs>
  auto zot(Expr&& expr, QuadratureArgs&&... args)
  {
    return makeOperator(tag::test_trial{}, FWD(expr), FWD(args)...);
  }

  /** @} **/

} // end namespace AMDiS
