#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec_trialvec {};
  }


  /// zero-order operator \f$ \langle\Psi, c\,\Phi\rangle \f$, or \f$ \langle\Psi, A\,\Phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::testvec_trialvec, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec_trialvec, LC, GridFct>, LC, GridFct>
  {
    static const int dow = ContextGeometry<LC>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    using expr_value_type = typename GridFct::Range;
    static_assert( Size_v<expr_value_type> == 1 || (Rows_v<expr_value_type> == dow && Cols_v<expr_value_type> == dow),
      "Expression must be of scalar or matrix type." );

  public:
    GridFunctionOperator(tag::testvec_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isPower && CN::isPower,
        "Operator can be applied to Power-Nodes only.");

      static_assert( (RN::CHILDREN == CN::CHILDREN), "" );
      // theoretically the restriction of the same nr of childs would not be necessary

      const bool sameFE = std::is_same<FiniteElementType_t<RN>, FiniteElementType_t<CN>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      using Category = ValueCategory_t<expr_value_type>;

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      if (sameFE && sameNode && std::is_same<Category,tag::scalar>::value)
        getElementMatrixOptimized(contextGeo, quad, rowNode, colNode, elementMatrix, Category{});
      else
        getElementMatrixStandard(contextGeo, quad, rowNode, colNode, elementMatrix, Category{});
    }

  protected: // specialization for different value-categories

    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixStandard(CG const& contextGeo, QR const& quad,
                                  RN const& rowNode, CN const& colNode,
                                  Mat& elementMatrix, tag::scalar)
    {
      static const std::size_t CHILDREN = RN::CHILDREN;

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      NodeQuadCache<typename RN::ChildType> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename CN::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value0 = value * colShapeValues[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              const auto local_kj = colNode.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
            }
          }
        }
      }
    }


    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixOptimized(CG const& contextGeo, QR const& quad,
                                   RN const& node, CN const& /*colNode*/,
                                   Mat& elementMatrix, tag::scalar)
    {
      static const std::size_t CHILDREN = RN::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();

      NodeQuadCache<typename RN::ChildType> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto value = factor * shapeValues[i];

          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_ki = node.child(k).localIndex(i);
            elementMatrix[local_ki][local_ki] += value * shapeValues[i];
          }

          for (std::size_t j = i+1; j < localFE.size(); ++j) {
            const auto value0 = value * shapeValues[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = node.child(k).localIndex(i);
              const auto local_kj = node.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
              elementMatrix[local_kj][local_ki] += value0;
            }
          }
        }
      }
    }

    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixStandard(CG const& contextGeo, QR const& quad,
                                  RN const& rowNode, CN const& colNode,
                                  Mat& elementMatrix, tag::matrix)
    {
      static const std::size_t CHILDREN = RN::CHILDREN;
      static_assert(Rows_v<expr_value_type> == CHILDREN && Cols_v<expr_value_type> == CHILDREN, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      NodeQuadCache<typename RN::ChildType> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename CN::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(contextGeo,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto exprValue = Super::coefficient(local);

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value0 = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value = value0 * colShapeValues[j];
            const auto mat = exprValue * value;

            for (std::size_t k0 = 0; k0 < CHILDREN; ++k0) {
              const auto local_ki = rowNode.child(k0).localIndex(i);
              for (std::size_t k1 = 0; k1 < CHILDREN; ++k1) {
                const auto local_kj = colNode.child(k1).localIndex(j);

                elementMatrix[local_ki][local_kj] += mat[k0][k1];
              }
            }
          }
        }
      }
    }

    template <class CG, class QR, class RN, class CN, class Mat>
    void getElementMatrixOptimized(CG const& contextGeo, QR const& quad,
                                   RN const& node, CN const& /*colNode*/,
                                   Mat& elementMatrix, tag::matrix)
    {
      assert(false && "Not implemented.");
    }
  };

  /** @} **/

} // end namespace AMDiS
