#pragma once

#include <dune/common/hybridutilities.hh>
#include <dune/common/std/type_traits.hh>
#include <dune/geometry/quadraturerules.hh>
#include <dune/geometry/type.hh>

#include <amdis/gridfunctions/Order.hpp>

namespace AMDiS
{
  /// Base class for quadrature factories for localFunctions
  template <class ctype, int dim, class LocalFunction>
  class QuadratureFactory
  {
  protected:
    using QuadratureRule = Dune::QuadratureRule<ctype, dim>;

  public:
    virtual ~QuadratureFactory() = default;

    /// Bind the rule to a localFunction
    virtual void bind(LocalFunction const& localFct)
    {}

    /// Return the quadrature order of the coefficient relates to the localFunction
    virtual int order() const
    {
      return 0;
    }

    /// Construct a quadrature rule, based on the local geometry-type and the
    /// polynomial degree of the integrand function.
    virtual QuadratureRule const& rule(Dune::GeometryType const& type, int degree) const
    {
      using QuadratureRules = Dune::QuadratureRules<ctype, dim>;
      return QuadratureRules::rule(type, degree);
    }
  };


  template <class LF>
  using HasLocalFunctionOrder = decltype( order(std::declval<LF>()) );


  /// \brief Factory for quadrature rule, that calculates the coefficient order from
  /// a localFunction passed to the bind method.
  template <class ctype, int dim, class LocalFunction>
  class QuadFactoryFromLocalFunction
      : public QuadratureFactory<ctype, dim, LocalFunction>
  {
    using Concept = Dune::Std::is_detected<HasLocalFunctionOrder, LocalFunction>;
    static_assert(Concept::value,
      "Polynomial order of GridFunction can not be extracted. Provide an explicit order parameter instead.");

  public:
    void bind(LocalFunction const& localFct) final
    {
      order_ = Dune::Hybrid::ifElse(Concept{},
        [&](auto id) -> int { return AMDiS::order(id(localFct)); },
        [] (auto)    -> int { return -1; });
    }

    int order() const final { return order_; }

  private:
    int order_ = -1;
  };

  struct PreQuadFactoryFromLocalFunction {};

  inline auto makePreQuadratureFactory()
  {
    return PreQuadFactoryFromLocalFunction{};
  }

  /// Generator for \ref QuadFactoryFromLocalFunction. \relates QuadFactoryFromLocalFunction
  template <class ctype, int dim, class LocalFunction>
  auto makeQuadratureFactory(PreQuadFactoryFromLocalFunction const& /*pre*/)
  {
    return QuadFactoryFromLocalFunction<ctype, dim, LocalFunction>{};
  }


  /// \brief Factory for quadrature rule, that takes to order of the localFunction
  /// and a quadrature-type
  template <class ctype, int dim, class LocalFunction>
  class QuadFactoryFromOrder
      : public QuadratureFactory<ctype, dim, LocalFunction>
  {
    using Super = QuadratureFactory<ctype, dim, LocalFunction>;
    using QuadratureRule = typename Super::QuadratureRule;

  public:
    /// Constructor. Takes the order of the localCoefficient and a quadrature-type
    QuadFactoryFromOrder(int order, Dune::QuadratureType::Enum qt)
      : order_(order)
      , qt_(qt)
    {}

    int order() const final { return order_; }

    QuadratureRule const& rule(Dune::GeometryType const& type, int degree) const final
    {
      using QuadratureRules = Dune::QuadratureRules<ctype, dim>;
      return QuadratureRules::rule(type, degree, qt_);
    }

  private:
    int order_;
    Dune::QuadratureType::Enum qt_;
  };

  struct PreQuadFactoryFromOrder
  {
    int order;
    Dune::QuadratureType::Enum qt;
  };

  inline auto makePreQuadratureFactory(int order, Dune::QuadratureType::Enum qt = Dune::QuadratureType::GaussLegendre)
  {
    return PreQuadFactoryFromOrder{order, qt};
  }

  /// Generator for \ref QuadFactoryFromOrder. \relates QuadFactoryFromOrder
  template <class ctype, int dim, class LocalFunction>
  auto makeQuadratureFactory(PreQuadFactoryFromOrder const& pre)
  {
    return QuadFactoryFromOrder<ctype, dim, LocalFunction>{pre.order, pre.qt};
  }


  /// \brief Factory for quadrature rule, that is based on an existing rule
  template <class ctype, int dim, class LocalFunction>
  class QuadFactoryFromRule
      : public QuadratureFactory<ctype, dim, LocalFunction>
  {
    using Super = QuadratureFactory<ctype, dim, LocalFunction>;
    using QuadratureRule = typename Super::QuadratureRule;

  public:
    QuadFactoryFromRule(QuadratureRule const& rule)
      : rule_(rule)
    {}

    QuadratureRule const& rule(Dune::GeometryType const& /*type*/, int /*degree*/) const final
    {
      return rule_;
    }

  private:
    QuadratureRule const& rule_;
  };

  template <class QuadRule>
  struct PreQuadFactoryFromRule
  {
    QuadRule const& rule;
  };

  template <class ctype, int dim>
  auto makePreQuadratureFactory(Dune::QuadratureRule<ctype, dim> const& rule)
  {
    return PreQuadFactoryFromRule<Dune::QuadratureRule<ctype, dim>>{rule};
  }

  /// Generator for \ref QuadFactoryFromRule. \relates QuadFactoryFromRule
  template <class ctype, int dim, class LocalFunction, class QuadRule>
  auto makeQuadratureFactory(PreQuadFactoryFromRule<QuadRule> const& pre)
  {
    return QuadFactoryFromRule<ctype, dim, LocalFunction>{pre.rule};
  }

} // end namespace AMDiS
